# Orca
#
# Copyright 2025 Valve Corporation
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-lines

"""Integration tests for Orca's D-Bus service."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

import pytest

from dasbus.error import DBusError
from gi.repository import GLib

if TYPE_CHECKING:
    from collections.abc import Callable
    from dasbus.client.proxy import ObjectProxy as DBusProxy

# Get configurable timeout values from environment variables
DEFAULT_MODULE_TIMEOUT = 15
FLAT_REVIEW_TIMEOUT = int(os.environ.get("ORCA_FLAT_REVIEW_TIMEOUT", DEFAULT_MODULE_TIMEOUT))
STRUCTURAL_NAVIGATOR_TIMEOUT = int(
    os.environ.get("ORCA_STRUCTURAL_NAVIGATOR_TIMEOUT", DEFAULT_MODULE_TIMEOUT)
)
SPEECH_AND_VERBOSITY_TIMEOUT = int(os.environ.get("ORCA_SPEECH_AND_VERBOSITY_TIMEOUT", 30))

MODULE_TIMEOUTS = {
    "FlatReviewPresenter": FLAT_REVIEW_TIMEOUT,
    "StructuralNavigator": STRUCTURAL_NAVIGATOR_TIMEOUT,
    "SpeechAndVerbosityManager": SPEECH_AND_VERBOSITY_TIMEOUT,
}

# Modules that may not be present in all environments (e.g., X11 vs Wayland)
OPTIONAL_MODULES = {"MouseReviewer"}

MODULE_CONFIG = {
    "ActionPresenter": {
        "commands": ["ShowActionsList"],
        "parameterized_commands": [],
        "getters": [],
        "setters": [],
        "ui_commands": ["ShowActionsList"],
        "toggle_commands": [],
        "skip": [],
    },
    "ClipboardPresenter": {
        "commands": ["PresentClipboardContents"],
        "parameterized_commands": [],
        "getters": [],
        "setters": [],
        "ui_commands": [],
        "toggle_commands": [],
        "skip": [],
    },
    "FocusManager": {
        "commands": [
            "EnableStickyBrowseMode",
            "EnableStickyFocusMode",
            "ToggleLayoutMode",
            "TogglePresentationMode",
        ],
        "parameterized_commands": [],
        "getters": [
            "BrowseModeIsSticky",
            "FocusModeIsSticky",
            "InFocusMode",
            "InLayoutMode",
        ],
        "setters": [],
        "ui_commands": [],
        "toggle_commands": ["ToggleLayoutMode", "TogglePresentationMode"],
        "skip": [],
    },
    "FlatReviewPresenter": {
        "commands": [
            "AppendToClipboard",
            "CopyToClipboard",
            "GetCurrentObject",
            "GoAbove",
            "GoBelow",
            "GoBottomLeft",
            "GoEnd",
            "GoEndOfLine",
            "GoHome",
            "GoNextCharacter",
            "GoNextItem",
            "GoNextLine",
            "GoPreviousCharacter",
            "GoPreviousItem",
            "GoPreviousLine",
            "GoStartOfLine",
            "LeftClickOnObject",
            "PhoneticItem",
            "PhoneticLine",
            "PresentCharacter",
            "PresentItem",
            "PresentLine",
            "PresentObject",
            "RightClickOnObject",
            "RoutePointerToObject",
            "SayAll",
            "ShowContents",
            "SpellCharacter",
            "SpellItem",
            "SpellLine",
            "ToggleFlatReviewMode",
            "ToggleRestrict",
            "UnicodeCurrentCharacter",
        ],
        "parameterized_commands": [],
        "getters": ["IsRestricted"],
        "setters": ["IsRestricted"],
        "ui_commands": [
            "ShowContents",
            "LeftClickOnObject",
            "RightClickOnObject",
            "RoutePointerToObject",
        ],
        "toggle_commands": ["ToggleFlatReviewMode", "ToggleRestrict"],
        "skip": [],
    },
    "ObjectNavigator": {
        "commands": [
            "MoveToFirstChild",
            "MoveToNextSibling",
            "MoveToParent",
            "MoveToPreviousSibling",
            "PerformAction",
            "ToggleSimplify",
        ],
        "parameterized_commands": [],
        "getters": [],
        "setters": [],
        "ui_commands": ["PerformAction"],
        "toggle_commands": ["ToggleSimplify"],
        "skip": [],
    },
    "NotificationPresenter": {
        "commands": [
            "PresentLastNotification",
            "PresentNextNotification",
            "PresentPreviousNotification",
            "ShowNotificationList",
        ],
        "parameterized_commands": [],
        "getters": [],
        "setters": [],
        "ui_commands": ["ShowNotificationList"],
        "toggle_commands": [],
        "skip": [],
    },
    "SleepModeManager": {
        "commands": ["ToggleSleepMode"],
        "parameterized_commands": [],
        "getters": [],
        "setters": [],
        "ui_commands": [],
        "toggle_commands": ["ToggleSleepMode"],
        "skip": [],
    },
    "SpeechAndVerbosityManager": {
        "commands": [
            "ChangeNumberStyle",
            "CycleCapitalizationStyle",
            "CyclePunctuationLevel",
            "CycleSynthesizer",
            "DecreasePitch",
            "DecreaseRate",
            "DecreaseVolume",
            "IncreasePitch",
            "IncreaseRate",
            "IncreaseVolume",
            "InterruptSpeech",
            "RefreshSpeech",
            "ShutdownSpeech",
            "StartSpeech",
            "ToggleIndentationAndJustification",
            "ToggleSpeech",
            "ToggleTableCellReadingMode",
            "ToggleVerbosity",
        ],
        "parameterized_commands": ["GetVoicesForLanguage"],
        "getters": [
            "AlwaysAnnounceSelectedRangeInSpreadsheet",
            "AnnounceBlockquote",
            "AnnounceCellCoordinates",
            "AnnounceCellHeaders",
            "AnnounceCellSpan",
            "AnnounceForm",
            "AnnounceGrouping",
            "AnnounceLandmark",
            "AnnounceList",
            "AnnounceSpreadsheetCellCoordinates",
            "AnnounceTable",
            "AvailableServers",
            "AvailableSynthesizers",
            "AvailableVoices",
            "CapitalizationStyle",
            "CurrentServer",
            "CurrentSynthesizer",
            "CurrentVoice",
            "InsertPausesBetweenUtterances",
            "MessagesAreDetailed",
            "OnlySpeakDisplayedText",
            "Pitch",
            "PunctuationLevel",
            "Rate",
            "RepeatedCharacterLimit",
            "SpeakBlankLines",
            "SpeakDescription",
            "SpeakIndentationAndJustification",
            "SpeakIndentationOnlyIfChanged",
            "SpeakMisspelledIndicator",
            "SpeakNumbersAsDigits",
            "SpeakPositionInSet",
            "SpeakRowInDocumentTable",
            "SpeakRowInGuiTable",
            "SpeakRowInSpreadsheet",
            "SpeakTutorialMessages",
            "SpeakWidgetMnemonic",
            "SpeechIsEnabled",
            "SpeechIsMuted",
            "UseColorNames",
            "UsePronunciationDictionary",
            "VerbosityLevel",
            "Volume",
        ],
        "setters": [
            "AlwaysAnnounceSelectedRangeInSpreadsheet",
            "AnnounceBlockquote",
            "AnnounceCellCoordinates",
            "AnnounceCellHeaders",
            "AnnounceCellSpan",
            "AnnounceForm",
            "AnnounceGrouping",
            "AnnounceLandmark",
            "AnnounceList",
            "AnnounceSpreadsheetCellCoordinates",
            "AnnounceTable",
            "CapitalizationStyle",
            "CurrentServer",
            "CurrentSynthesizer",
            "CurrentVoice",
            "InsertPausesBetweenUtterances",
            "MessagesAreDetailed",
            "OnlySpeakDisplayedText",
            "Pitch",
            "PunctuationLevel",
            "Rate",
            "RepeatedCharacterLimit",
            "SpeakBlankLines",
            "SpeakDescription",
            "SpeakIndentationAndJustification",
            "SpeakIndentationOnlyIfChanged",
            "SpeakMisspelledIndicator",
            "SpeakNumbersAsDigits",
            "SpeakPositionInSet",
            "SpeakRowInDocumentTable",
            "SpeakRowInGuiTable",
            "SpeakRowInSpreadsheet",
            "SpeakTutorialMessages",
            "SpeakWidgetMnemonic",
            "SpeechIsEnabled",
            "SpeechIsMuted",
            "UseColorNames",
            "UsePronunciationDictionary",
            "VerbosityLevel",
            "Volume",
        ],
        "ui_commands": [],
        "toggle_commands": [
            "ToggleIndentationAndJustification",
            "ToggleSpeech",
            "ToggleTableCellReadingMode",
            "ToggleVerbosity",
        ],
        "skip": [],
    },
    "StructuralNavigator": {
        "commands": [
            "ContainerEnd",
            "ContainerStart",
            "CycleMode",
            "ListBlockquotes",
            "ListButtons",
            "ListCheckboxes",
            "ListClickables",
            "ListComboboxes",
            "ListEntries",
            "ListFormFields",
            "ListHeadings",
            "ListHeadingsLevel1",
            "ListHeadingsLevel2",
            "ListHeadingsLevel3",
            "ListHeadingsLevel4",
            "ListHeadingsLevel5",
            "ListHeadingsLevel6",
            "ListIframes",
            "ListImages",
            "ListLandmarks",
            "ListLargeObjects",
            "ListLinks",
            "ListListItems",
            "ListLists",
            "ListParagraphs",
            "ListRadioButtons",
            "ListTables",
            "ListUnvisitedLinks",
            "ListVisitedLinks",
            "NextBlockquote",
            "NextButton",
            "NextCheckbox",
            "NextClickable",
            "NextCombobox",
            "NextEntry",
            "NextFormField",
            "NextHeading",
            "NextHeadingLevel1",
            "NextHeadingLevel2",
            "NextHeadingLevel3",
            "NextHeadingLevel4",
            "NextHeadingLevel5",
            "NextHeadingLevel6",
            "NextIframe",
            "NextImage",
            "NextLandmark",
            "NextLargeObject",
            "NextLink",
            "NextList",
            "NextListItem",
            "NextLiveRegion",
            "NextParagraph",
            "NextRadioButton",
            "NextSeparator",
            "NextTable",
            "NextUnvisitedLink",
            "NextVisitedLink",
            "PreviousBlockquote",
            "PreviousButton",
            "PreviousCheckbox",
            "PreviousClickable",
            "PreviousCombobox",
            "PreviousEntry",
            "PreviousFormField",
            "PreviousHeading",
            "PreviousHeadingLevel1",
            "PreviousHeadingLevel2",
            "PreviousHeadingLevel3",
            "PreviousHeadingLevel4",
            "PreviousHeadingLevel5",
            "PreviousHeadingLevel6",
            "PreviousIframe",
            "PreviousImage",
            "PreviousLandmark",
            "PreviousLargeObject",
            "PreviousLink",
            "PreviousList",
            "PreviousListItem",
            "PreviousLiveRegion",
            "PreviousParagraph",
            "PreviousRadioButton",
            "PreviousSeparator",
            "PreviousTable",
            "PreviousUnvisitedLink",
            "PreviousVisitedLink",
        ],
        "parameterized_commands": [],
        "getters": [
            "IsEnabled",
            "LargeObjectTextLength",
            "NavigationWraps",
            "TriggersFocusMode",
        ],
        "setters": [
            "IsEnabled",
            "LargeObjectTextLength",
            "NavigationWraps",
            "TriggersFocusMode",
        ],
        "ui_commands": [
            "ListBlockquotes",
            "ListButtons",
            "ListCheckboxes",
            "ListClickables",
            "ListComboboxes",
            "ListEntries",
            "ListFormFields",
            "ListHeadings",
            "ListHeadingsLevel1",
            "ListHeadingsLevel2",
            "ListHeadingsLevel3",
            "ListHeadingsLevel4",
            "ListHeadingsLevel5",
            "ListHeadingsLevel6",
            "ListIframes",
            "ListImages",
            "ListLandmarks",
            "ListLargeObjects",
            "ListLinks",
            "ListListItems",
            "ListLists",
            "ListParagraphs",
            "ListRadioButtons",
            "ListTables",
            "ListUnvisitedLinks",
            "ListVisitedLinks",
        ],
        "toggle_commands": [],
        "skip": ["CycleMode"],  # Test flakiness depending on app and focus.
    },
    "SystemInformationPresenter": {
        "commands": [
            "PresentBatteryStatus",
            "PresentCpuAndMemoryUsage",
            "PresentDate",
            "PresentTime",
        ],
        "parameterized_commands": [],
        "getters": [
            "DateFormat",
            "TimeFormat",
            "AvailableDateFormats",
            "AvailableTimeFormats",
        ],
        "setters": [
            "DateFormat",
            "TimeFormat",
        ],
        "ui_commands": [],
        "toggle_commands": [],
        "skip": ["PresentBatteryStatus"],  # Can timeout on systems without battery
    },
    "TableNavigator": {
        "commands": [
            "ClearDynamicColumnHeadersRow",
            "ClearDynamicRowHeadersColumn",
            "MoveDown",
            "MoveLeft",
            "MoveRight",
            "MoveToBeginningOfRow",
            "MoveToBottomOfColumn",
            "MoveToEndOfRow",
            "MoveToFirstCell",
            "MoveToLastCell",
            "MoveToTopOfColumn",
            "MoveUp",
            "SetDynamicColumnHeadersRow",
            "SetDynamicRowHeadersColumn",
            "ToggleEnabled",
        ],
        "parameterized_commands": [],
        "getters": ["SkipBlankCells"],
        "setters": ["SkipBlankCells"],
        "ui_commands": [],
        "toggle_commands": ["ToggleEnabled"],
        "skip": [],
    },
    "TypingEchoPresenter": {
        "commands": ["CycleKeyEcho"],
        "parameterized_commands": [],
        "getters": [
            "KeyEchoEnabled",
            "CharacterEchoEnabled",
            "WordEchoEnabled",
            "SentenceEchoEnabled",
            "AlphabeticKeysEnabled",
            "NumericKeysEnabled",
            "PunctuationKeysEnabled",
            "SpaceEnabled",
            "ModifierKeysEnabled",
            "FunctionKeysEnabled",
            "ActionKeysEnabled",
            "NavigationKeysEnabled",
            "DiacriticalKeysEnabled",
            "LockingKeysPresented",
        ],
        "setters": [
            "KeyEchoEnabled",
            "CharacterEchoEnabled",
            "WordEchoEnabled",
            "SentenceEchoEnabled",
            "AlphabeticKeysEnabled",
            "NumericKeysEnabled",
            "PunctuationKeysEnabled",
            "SpaceEnabled",
            "ModifierKeysEnabled",
            "FunctionKeysEnabled",
            "ActionKeysEnabled",
            "NavigationKeysEnabled",
            "DiacriticalKeysEnabled",
            "LockingKeysPresented",
        ],
        "ui_commands": [],
        "toggle_commands": [],
        "skip": [],
    },
    "CaretNavigator": {
        "commands": [
            "ToggleEnabled",
            "NextCharacter",
            "PreviousCharacter",
            "NextWord",
            "PreviousWord",
            "NextLine",
            "PreviousLine",
            "StartOfLine",
            "EndOfLine",
            "StartOfFile",
            "EndOfFile",
        ],
        "parameterized_commands": [],
        "getters": ["IsEnabled", "TriggersFocusMode"],
        "setters": ["IsEnabled", "TriggersFocusMode"],
        "ui_commands": [],
        "toggle_commands": ["ToggleEnabled"],
        "skip": [],
    },
    "SayAllPresenter": {
        "commands": ["SayAll", "Rewind", "FastForward"],
        "parameterized_commands": [],
        "getters": [
            "AnnounceBlockquote",
            "AnnounceForm",
            "AnnounceGrouping",
            "AnnounceLandmark",
            "AnnounceList",
            "AnnounceTable",
            "Style",
            "StructuralNavigationEnabled",
            "RewindAndFastForwardEnabled",
        ],
        "setters": [
            "AnnounceBlockquote",
            "AnnounceForm",
            "AnnounceGrouping",
            "AnnounceLandmark",
            "AnnounceList",
            "AnnounceTable",
            "Style",
            "StructuralNavigationEnabled",
            "RewindAndFastForwardEnabled",
        ],
        "ui_commands": [],
        "toggle_commands": [],
        "skip": [],
    },
    "WhereAmIPresenter": {
        "commands": [
            "PresentCharacterAttributes",
            "PresentDefaultButton",
            "PresentLink",
            "PresentSelectedText",
            "PresentSelection",
            "PresentSizeAndPosition",
            "PresentStatusBar",
            "PresentTitle",
            "WhereAmIBasic",
            "WhereAmIDetailed",
        ],
        "parameterized_commands": [],
        "getters": [],
        "setters": [],
        "ui_commands": [],
        "toggle_commands": [],
        "skip": [],
    },
    "BraillePresenter": {
        "commands": [],
        "parameterized_commands": [],
        "getters": [
            "BrailleIsEnabled",
            "VerbosityLevel",
            "RolenameStyle",
            "DisplayAncestors",
            "ContractedBrailleIsEnabled",
            "ContractionTable",
            "AvailableContractionTables",
            "EndOfLineIndicatorIsEnabled",
            "WordWrapIsEnabled",
            "FlashMessagesAreEnabled",
            "FlashMessageDuration",
            "FlashMessagesArePersistent",
            "FlashMessagesAreDetailed",
            "SelectorIndicator",
            "LinkIndicator",
            "TextAttributesIndicator",
        ],
        "setters": [
            "BrailleIsEnabled",
            "VerbosityLevel",
            "RolenameStyle",
            "DisplayAncestors",
            "ContractedBrailleIsEnabled",
            "ContractionTable",
            "EndOfLineIndicatorIsEnabled",
            "WordWrapIsEnabled",
            "FlashMessagesAreEnabled",
            "FlashMessageDuration",
            "FlashMessagesArePersistent",
            "FlashMessagesAreDetailed",
            "SelectorIndicator",
            "LinkIndicator",
            "TextAttributesIndicator",
        ],
        "ui_commands": [],
        "toggle_commands": [],
        "skip": [],
    },
}

PARAMETERIZED_TEST_PARAMS = {
    "GetVoicesForLanguage": {"language": "en", "variant": "", "notify_user": False}
}


def safe_call(func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any | None:
    """Safely call a function, returning result or None on error."""

    try:
        return func(*args, **kwargs)
    except (DBusError, AttributeError, TypeError, ValueError):
        return None


def extract_names(items: list[Any] | None) -> list[str]:
    """Extract first element from list of tuples/lists, return empty list if None."""

    return [item[0] for item in items] if items else []


def is_context_error(error_str: str) -> bool:
    """Check if error indicates missing context (script/runtime).

    This function checks for common error patterns that indicate the Orca
    service is running but lacks the proper context (e.g., no active script
    or application context) to execute the requested operation.
    """

    context_error_patterns = [
        # NoneType attribute errors indicate missing objects/context
        ("NoneType" in error_str and "attribute" in error_str),
        # Script-related None errors indicate missing script context
        ("script" in error_str and "None" in error_str),
        # AT-SPI context missing
        ("context" in error_str.lower() and "none" in error_str.lower()),
        # Missing application object
        ("application" in error_str.lower() and "none" in error_str.lower()),
    ]

    return any(pattern for pattern in context_error_patterns)


def is_timeout_error(error_str: str) -> bool:
    """Check if error indicates a timeout condition."""

    timeout_patterns = [
        "timed out" in error_str.lower(),
        "timeout" in error_str.lower(),
        "deadline exceeded" in error_str.lower(),
        "connection timeout" in error_str.lower(),
    ]

    return any(pattern for pattern in timeout_patterns)


def unpack_variant(value: Any) -> Any:
    """Unpack a GLib.Variant if needed."""

    return value.unpack() if hasattr(value, "unpack") else value


def get_alternative_value(
    proxy: DBusProxy, prop_name: str, current_value: str | int | float | bool
) -> str | int | float | bool:
    """Get an alternative value for testing setters, avoiding defaults when possible."""

    try:
        if prop_name == "CurrentServer":
            servers = proxy.ExecuteRuntimeGetter("AvailableServers")
            servers = unpack_variant(servers)
            assert isinstance(current_value, str)
            non_default = [s for s in servers if s != current_value and "default" not in s.lower()]
            if non_default:
                return non_default[0]
            return next((s for s in servers if s != current_value), current_value)
        if prop_name == "CurrentSynthesizer":
            synths = proxy.ExecuteRuntimeGetter("AvailableSynthesizers")
            synths = unpack_variant(synths)
            assert isinstance(current_value, str)
            non_default = [s for s in synths if s != current_value and "default" not in s.lower()]
            if non_default:
                return non_default[0]
            return next((s for s in synths if s != current_value), current_value)
        if prop_name == "CurrentVoice":
            voices = proxy.ExecuteRuntimeGetter("AvailableVoices")
            voices = unpack_variant(voices)
            if voices and len(voices) > 1:
                assert isinstance(current_value, str)
                current_voice_name = current_value.split()[0] if current_value else ""
                non_default = [
                    v[0]
                    for v in voices
                    if v[0] != current_voice_name and "default" not in v[0].lower()
                ]
                if non_default:
                    return non_default[0]
                return next((v[0] for v in voices if v[0] != current_voice_name), current_value)
            return current_value
        return current_value
    except (DBusError, AttributeError, TypeError, ValueError):
        return current_value


def get_test_value(
    proxy: DBusProxy, prop_name: str, current_value: str | int | float | bool
) -> str | int | float | bool:
    """Generate an appropriate test value for a property."""

    if isinstance(current_value, (int, float)):
        return current_value + 1 if current_value < 100 else current_value - 1
    if isinstance(current_value, str):
        return get_alternative_value(proxy, prop_name, current_value)
    if isinstance(current_value, bool):
        return not current_value
    return current_value


def to_variant(value: str | bool | int | float) -> Any:
    """Convert a Python value to GLib.Variant."""

    if isinstance(value, str):
        return GLib.Variant("s", value)
    if isinstance(value, bool):
        return GLib.Variant("b", value)
    if isinstance(value, int):
        return GLib.Variant("i", value)
    if isinstance(value, float):
        return GLib.Variant("d", value)
    return GLib.Variant("s", str(value))


@pytest.mark.dbus
class TestOrcaDBusIntegration:
    """Integration tests for Orca D-Bus service using pytest features."""

    @pytest.mark.dbus
    def test_orca_dbus_service_running(self, dbus_service_proxy):
        """Test that Orca D-Bus service is running and accessible."""

        version = dbus_service_proxy.GetVersion()
        assert version is not None
        assert len(str(version)) > 0

    @pytest.mark.dbus
    def test_list_modules(self, dbus_service_proxy):
        """Test listing available modules."""

        modules = list(dbus_service_proxy.ListModules())
        assert isinstance(modules, list)

        expected_modules = set(MODULE_CONFIG.keys())
        actual_modules = set(modules)
        missing_modules = expected_modules - actual_modules

        assert not missing_modules, f"Missing expected modules: {missing_modules}"

    @pytest.mark.dbus
    def test_present_message(self, dbus_service_proxy):
        """Test the PresentMessage service command."""

        dbus_service_proxy.PresentMessage("Integration test message")

        # TODO - JD: What we really want to do is verify that the message was presented.
        # That will require we have a way to capture the speech and braille. Until then,
        # verify service is still responsive after presentation.
        version = dbus_service_proxy.GetVersion()
        assert version is not None

    @pytest.mark.dbus
    def test_error_handling(self, dbus_service_proxy):
        """Test error handling for invalid D-Bus calls."""

        with pytest.raises((DBusError, AttributeError, TypeError)):
            dbus_service_proxy.NonExistentMethod()

    @pytest.mark.parametrize(
        "module_name,config", MODULE_CONFIG.items(), ids=list(MODULE_CONFIG.keys())
    )
    @pytest.mark.dbus
    def test_module_capabilities(self, module_proxy_factory, run_with_timeout, module_name, config):
        """Test that each module reports correct capabilities."""
        print(f"\n  Testing {module_name} capabilities:")
        for cap_type in ["commands", "parameterized_commands", "getters", "setters"]:
            items = config.get(cap_type, [])
            if items:
                print(f"    • {cap_type}: {len(items)} items")
                for item in sorted(items)[:5]:  # Show first 5 items
                    print(f"      - {item}")
                if len(items) > 5:
                    print(f"      - ... and {len(items) - 5} more")

        def get_capabilities():
            proxy = module_proxy_factory(module_name)
            return {
                "commands": extract_names(safe_call(proxy.ListCommands)),
                "parameterized_commands": extract_names(safe_call(proxy.ListParameterizedCommands)),
                "getters": extract_names(safe_call(proxy.ListRuntimeGetters)),
                "setters": extract_names(safe_call(proxy.ListRuntimeSetters)),
            }

        timeout = MODULE_TIMEOUTS.get(module_name)
        result = run_with_timeout(get_capabilities, timeout)
        assert result["success"], f"Failed to get {module_name} capabilities: {result['error']}"

        for cap_type in ["commands", "parameterized_commands", "getters", "setters"]:
            expected = set(config.get(cap_type, []))
            actual = set(result["result"].get(cap_type, []))
            missing = expected - actual
            unexpected = actual - expected

            assert not missing, f"{module_name} missing {cap_type}: {sorted(missing)}"
            assert not unexpected, f"{module_name} unexpected {cap_type}: {sorted(unexpected)}"

    @pytest.mark.parametrize(
        "module_name,config",
        [(name, config) for name, config in MODULE_CONFIG.items() if config.get("commands")],
        ids=[name for name, config in MODULE_CONFIG.items() if config.get("commands")],
    )
    @pytest.mark.dbus
    def test_module_commands(self, module_proxy_factory, run_with_timeout, module_name, config):
        """Test that module commands execute without errors."""
        commands = config["commands"]
        ui_commands = config.get("ui_commands", [])
        toggle_commands = config.get("toggle_commands", [])
        skip_commands = config.get("skip", [])
        print(f"\n  Testing {module_name} commands ({len(commands)} total):")
        for cmd in sorted(commands):
            status_parts = []
            if cmd in ui_commands:
                status_parts.append("UI - skipped")
            if cmd in skip_commands:
                status_parts.append("skipped")
            if cmd in toggle_commands:
                status_parts.append("toggle - restore state")
            status = f"({', '.join(status_parts)})" if status_parts else ""
            print(f"    • {cmd} {status}")

        def test_single_command(proxy, cmd_name, ui_commands, toggle_commands, skip_commands):
            if cmd_name in ui_commands or cmd_name in skip_commands:
                return {"success": True, "skipped": True}
            try:
                proxy.ExecuteCommand(cmd_name, False)
                if cmd_name in toggle_commands:
                    print(f"      → Restoring {cmd_name} to original state")
                    proxy.ExecuteCommand(cmd_name, False)
                return {"success": True}
            except (DBusError, AttributeError, TypeError, ValueError) as error:
                error_str = str(error)
                if is_context_error(error_str) or is_timeout_error(error_str):
                    return {"success": True, "context_required": True}
                return {"success": False, "error": error_str}

        def test_commands():
            proxy = module_proxy_factory(module_name)
            ui_commands = config.get("ui_commands", [])
            toggle_commands = config.get("toggle_commands", [])
            skip_commands = config.get("skip", [])
            return {
                cmd: test_single_command(proxy, cmd, ui_commands, toggle_commands, skip_commands)
                for cmd in config["commands"]
            }

        timeout = MODULE_TIMEOUTS.get(module_name)
        result = run_with_timeout(test_commands, timeout)
        assert result["success"], f"Failed to test {module_name} commands: {result['error']}"

        failed = [
            f"{cmd}: {res['error']}"
            for cmd, res in result["result"].items()
            if not res["success"] and not res.get("skipped")
        ]
        assert not failed, f"{module_name} command failures: {failed}"

    @pytest.mark.parametrize(
        "module_name,config",
        [
            (name, config)
            for name, config in MODULE_CONFIG.items()
            if config.get("parameterized_commands")
        ],
        ids=[
            name for name, config in MODULE_CONFIG.items() if config.get("parameterized_commands")
        ],
    )
    @pytest.mark.dbus
    def test_module_parameterized_commands(
        self, module_proxy_factory, run_with_timeout, module_name, config
    ):
        """Test that module parameterized commands execute with proper parameters."""
        param_commands = config["parameterized_commands"]
        print(f"\n  Testing {module_name} parameterized commands ({len(param_commands)} total):")
        for cmd in sorted(param_commands):
            params = PARAMETERIZED_TEST_PARAMS.get(cmd, {})
            param_str = ", ".join(f"{k}={v}" for k, v in params.items())
            print(f"    • {cmd}({param_str})")

        def test_single_param_command(proxy, cmd_name):
            params = PARAMETERIZED_TEST_PARAMS.get(cmd_name, {})
            if not params:
                return {"success": False, "error": "No test parameters"}

            try:
                variant_params = {k: to_variant(v) for k, v in params.items() if k != "notify_user"}
                notify_user = params.get("notify_user", False)
                result = proxy.ExecuteParameterizedCommand(cmd_name, variant_params, notify_user)
                return {"success": True, "result": result}
            except (DBusError, AttributeError, TypeError, ValueError) as error:
                if is_context_error(str(error)):
                    return {"success": True, "context_required": True}
                return {"success": False, "error": str(error)}

        def test_parameterized_commands():
            if GLib is None:
                pytest.skip("GLib not available")
            proxy = module_proxy_factory(module_name)
            return {
                cmd: test_single_param_command(proxy, cmd)
                for cmd in config["parameterized_commands"]
            }

        timeout = MODULE_TIMEOUTS.get(module_name)
        result = run_with_timeout(test_parameterized_commands, timeout)
        error_msg = f"Failed to test {module_name} parameterized commands: {result['error']}"
        assert result["success"], error_msg

        failed = [
            f"{cmd}: {res['error']}"
            for cmd, res in result["result"].items()
            if not res["success"] and not res.get("context_required")
        ]
        assert not failed, f"{module_name} parameterized command failures: {failed}"

    @pytest.mark.parametrize(
        "module_name,config",
        [
            (name, config)
            for name, config in MODULE_CONFIG.items()
            if config.get("getters") or config.get("setters")
        ],
        ids=[
            name
            for name, config in MODULE_CONFIG.items()
            if config.get("getters") or config.get("setters")
        ],
    )
    @pytest.mark.dbus
    def test_module_getters_setters(
        self, module_proxy_factory, run_with_timeout, module_name, config
    ):
        """Test that module getter/setter pairs work correctly."""
        all_props = sorted(set(config.get("getters", []) + config.get("setters", [])))
        print(f"\n  Testing {module_name} properties ({len(all_props)} total):")
        for prop in all_props:
            has_getter = prop in config.get("getters", [])
            has_setter = prop in config.get("setters", [])
            status_parts = []
            if has_getter:
                status_parts.append("get")
            if has_setter:
                status_parts.append("set")
            status = f"({'/'.join(status_parts)})" if status_parts else ""
            print(f"    • {prop} {status}")

        def test_single_property(proxy, prop_name, is_setter=False):
            try:
                current_value = unpack_variant(proxy.ExecuteRuntimeGetter(prop_name))
                if is_setter:
                    test_value = get_test_value(proxy, prop_name, current_value)
                    proxy.ExecuteRuntimeSetter(prop_name, to_variant(test_value))
                    new_value = unpack_variant(proxy.ExecuteRuntimeGetter(prop_name))
                    proxy.ExecuteRuntimeSetter(prop_name, to_variant(current_value))
                    return {
                        "success": True,
                        "original_value": current_value,
                        "test_value": test_value,
                        "actual_new_value": new_value,
                    }
                return {"success": True, "value": current_value}
            except (DBusError, AttributeError, TypeError, ValueError) as error:
                return {"success": False, "error": str(error)}

        def test_getters_setters():
            proxy = module_proxy_factory(module_name)
            results = {}

            for prop in config.get("getters", []):
                results[f"get_{prop}"] = test_single_property(proxy, prop, is_setter=False)

            for prop in config.get("setters", []):
                results[f"set_{prop}"] = test_single_property(proxy, prop, is_setter=True)

            return results

        timeout = MODULE_TIMEOUTS.get(module_name)
        result = run_with_timeout(test_getters_setters, timeout)
        assert result["success"], f"Failed to test {module_name} getters/setters: {result['error']}"

        if result["result"]:
            getter_results = {k: v for k, v in result["result"].items() if k.startswith("get_")}
            setter_results = {k: v for k, v in result["result"].items() if k.startswith("set_")}

            if getter_results:
                print("    Getter values:")
                for prop_key, res in getter_results.items():
                    prop = prop_key[4:]
                    if res["success"]:
                        print(f"      - {prop}: {res['value']}")
                    else:
                        print(f"      - {prop}: ERROR - {res['error']}")

            if setter_results:
                print("    Setter tests:")
                for prop_key, res in setter_results.items():
                    prop = prop_key[4:]
                    if res["success"]:
                        original = res["original_value"]
                        test_val = res["test_value"]
                        actual = res["actual_new_value"]
                        print(f"      - {prop}: {original} → {test_val} (got: {actual})")
                    else:
                        print(f"      - {prop}: ERROR - {res['error']}")

        failed = [
            f"{prop}: {res['error']}"
            for prop, res in result["result"].items()
            if not res["success"]
        ]
        assert not failed, f"{module_name} getter/setter failures: {failed}"

    @pytest.mark.dbus
    def test_parameterized_command_signatures(self, module_proxy_factory, run_with_timeout):
        """Test that parameterized commands have correct parameter signatures."""

        def get_signatures():
            proxy = module_proxy_factory("SpeechAndVerbosityManager")
            commands = proxy.ListParameterizedCommands()
            return {cmd[0]: [list(p[:2]) for p in cmd[2]] for cmd in commands}

        result = run_with_timeout(get_signatures)
        error_msg = f"Could not get parameterized command signatures: {result['error']}"
        assert result["success"], error_msg

        signatures = result["result"]
        if "GetVoicesForLanguage" in signatures:
            expected = [["language", "str"], ["variant", "str"], ["notify_user", "bool"]]
            actual = signatures["GetVoicesForLanguage"]
            error_msg = (
                f"GetVoicesForLanguage signature mismatch: expected {expected}, got {actual}"
            )
            assert actual == expected, error_msg

    @pytest.mark.dbus
    def test_no_unexpected_modules(self, dbus_service_proxy):
        """Test that no unexpected modules exist - ensures test coverage for all modules."""

        actual_modules = set(dbus_service_proxy.ListModules())
        expected_modules = set(MODULE_CONFIG.keys())
        unexpected_modules = actual_modules - expected_modules - OPTIONAL_MODULES

        if unexpected_modules:
            module_list = sorted(unexpected_modules)
            error_msg = (
                f"Found {len(unexpected_modules)} unexpected modules that lack test coverage: "
                f"{module_list}\n"
                f"Please add configuration for these modules to MODULE_CONFIG in the test file."
            )
            assert False, error_msg



# Unit tests for orca_modifier_manager.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=protected-access
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-locals

"""Unit tests for orca_modifier_manager.py methods."""

from __future__ import annotations

import subprocess
from typing import TYPE_CHECKING
from unittest.mock import call

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext
    from unittest.mock import MagicMock

@pytest.mark.unit
class TestOrcaModifierManager:
    """Test OrcaModifierManager class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Returns dependencies for orca_modifier_manager module testing."""

        additional_modules = ["orca.input_event_manager", "gi.repository"]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        debug_mock = essential_modules["orca.debug"]
        debug_mock.print_message = test_context.Mock()
        debug_mock.LEVEL_INFO = 800

        keybindings_mock = essential_modules["orca.keybindings"]
        keybindings_mock.get_keycodes = test_context.Mock(return_value=(0, 0))

        input_event_manager_mock = essential_modules["orca.input_event_manager"]
        input_manager_instance = test_context.Mock()
        input_manager_instance.add_grab_for_modifier = test_context.Mock(return_value=123)
        input_manager_instance.remove_grab_for_modifier = test_context.Mock()
        input_event_manager_mock.get_manager = test_context.Mock(
            return_value=input_manager_instance
        )

        settings_manager_mock = essential_modules["orca.settings_manager"]
        settings_manager_instance = test_context.Mock()
        settings_manager_instance.get_setting = test_context.Mock(
            return_value=["Insert", "KP_Insert"]
        )
        settings_manager_mock.get_manager = test_context.Mock(
            return_value=settings_manager_instance
        )

        gi_repository_mock = essential_modules["gi.repository"]

        gdk_mock = test_context.Mock()
        display_mock = test_context.Mock()
        device_manager_mock = test_context.Mock()
        display_mock.get_device_manager = test_context.Mock(return_value=device_manager_mock)
        gdk_mock.Display = test_context.Mock()
        gdk_mock.Display.get_default = test_context.Mock(return_value=display_mock)
        gdk_mock.InputSource = test_context.Mock()
        gdk_mock.InputSource.KEYBOARD = 4
        gi_repository_mock.Gdk = gdk_mock

        atspi_mock = test_context.Mock()
        atspi_mock.generate_keyboard_event = test_context.Mock()
        atspi_mock.ModifierType = test_context.Mock()
        atspi_mock.ModifierType.SHIFTLOCK = 1
        atspi_mock.ModifierType.SHIFT = 0
        gi_repository_mock.Atspi = atspi_mock

        glib_mock = test_context.Mock()
        glib_mock.timeout_add = test_context.Mock()
        gi_repository_mock.GLib = glib_mock

        essential_modules["input_manager_instance"] = input_manager_instance
        essential_modules["settings_manager_instance"] = settings_manager_instance
        essential_modules["gdk"] = gdk_mock
        essential_modules["atspi"] = atspi_mock
        essential_modules["glib"] = glib_mock

        return essential_modules

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test OrcaModifierManager.__init__."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        mock_gdk = test_context.Mock()
        test_context.patch("orca.orca_modifier_manager.Gdk", new=mock_gdk)
        mock_display = test_context.Mock()
        mock_device_manager = test_context.Mock()
        mock_display.get_device_manager.return_value = mock_device_manager
        mock_gdk.Display.get_default.return_value = mock_display
        manager = orca_modifier_manager.OrcaModifierManager()
        assert not manager._grabbed_modifiers
        assert manager._is_pressed is False
        assert manager._original_xmodmap == b""
        assert manager._caps_lock_cleared is False
        assert manager._need_to_restore_orca_modifier is False

        assert manager is not None

    def test_init_no_display(self, test_context: OrcaTestContext) -> None:
        """Test OrcaModifierManager.__init__ with no display available."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        mock_gdk = test_context.Mock()
        test_context.patch("orca.orca_modifier_manager.Gdk", new=mock_gdk)
        mock_gdk.Display.get_default.return_value = None
        essential_modules["orca.debug"].reset_mock()
        manager = orca_modifier_manager.OrcaModifierManager()

        assert not manager._grabbed_modifiers
        assert manager._is_pressed is False
        assert manager._original_xmodmap == b""
        assert manager._caps_lock_cleared is False
        assert manager._need_to_restore_orca_modifier is False

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "keyboard_device",
                "device_source": 4,
                "should_refresh": True,
            },  # Gdk.InputSource.KEYBOARD = 4
            {
                "id": "mouse_device",
                "device_source": 0,
                "should_refresh": False,
            },  # Gdk.InputSource.MOUSE = 0
            {
                "id": "touchscreen_device",
                "device_source": 5,
                "should_refresh": False,
            },  # Gdk.InputSource.TOUCHSCREEN = 5
        ],
        ids=lambda case: case["id"],
    )
    def test_on_device_changed(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test OrcaModifierManager._on_device_changed."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()

        mock_device = test_context.Mock()
        mock_device.get_source.return_value = case["device_source"]

        mock_refresh = test_context.Mock()
        test_context.patch_object(manager, "refresh_orca_modifiers", new=mock_refresh)
        test_context.patch("orca.orca_modifier_manager.Gdk.InputSource.KEYBOARD", new=4)
        manager._on_device_changed(None, mock_device)
        if case["should_refresh"]:
            mock_refresh.assert_called_once_with("Keyboard change detected.")
        else:
            mock_refresh.assert_not_called()

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "insert_grabbed",
                "modifier": "Insert",
                "orca_modifier_keys": ["Insert", "KP_Insert"],
                "is_grabbed": True,
                "expected_result": True,
            },
            {
                "id": "insert_not_grabbed",
                "modifier": "Insert",
                "orca_modifier_keys": ["Insert", "KP_Insert"],
                "is_grabbed": False,
                "expected_result": False,
            },
            {
                "id": "kp_insert_grabbed",
                "modifier": "KP_Insert",
                "orca_modifier_keys": ["Insert", "KP_Insert"],
                "is_grabbed": True,
                "expected_result": True,
            },
            {
                "id": "kp_insert_not_grabbed",
                "modifier": "KP_Insert",
                "orca_modifier_keys": ["Insert", "KP_Insert"],
                "is_grabbed": False,
                "expected_result": False,
            },
            {
                "id": "caps_lock_always_true",
                "modifier": "Caps_Lock",
                "orca_modifier_keys": ["Caps_Lock"],
                "is_grabbed": False,
                "expected_result": True,
            },
            {
                "id": "shift_lock_always_true",
                "modifier": "Shift_Lock",
                "orca_modifier_keys": ["Shift_Lock"],
                "is_grabbed": False,
                "expected_result": True,
            },
            {
                "id": "not_orca_modifier",
                "modifier": "Control_L",
                "orca_modifier_keys": ["Insert"],
                "is_grabbed": False,
                "expected_result": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_is_orca_modifier(
        self,
        test_context,
        case: dict,
    ) -> None:
        """Test OrcaModifierManager.is_orca_modifier."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        manager._grabbed_modifiers = {"Insert": 1, "KP_Insert": 2} if case["is_grabbed"] else {}

        mock_sm = test_context.Mock()
        test_context.patch(
            "orca.orca_modifier_manager.settings_manager.get_manager", new=mock_sm
        )
        mock_sm.return_value.get_setting.return_value = case["orca_modifier_keys"]
        result = manager.is_orca_modifier(case["modifier"])
        assert result == case["expected_result"]

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "get_initial_false",
                "test_type": "get",
                "initial_state": False,
                "new_state": None,
                "expected_get": False,
                "expected_set": None,
            },
            {
                "id": "get_set_true",
                "test_type": "get",
                "initial_state": True,
                "new_state": None,
                "expected_get": True,
                "expected_set": None,
            },
            {
                "id": "set_to_true",
                "test_type": "set",
                "initial_state": None,
                "new_state": True,
                "expected_get": None,
                "expected_set": True,
            },
            {
                "id": "set_to_false",
                "test_type": "set",
                "initial_state": None,
                "new_state": False,
                "expected_get": None,
                "expected_set": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_pressed_state_operations(self, test_context, case: dict) -> None:
        """Test OrcaModifierManager get_pressed_state and set_pressed_state."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        essential_modules["orca.debug"].reset_mock()

        if case["test_type"] == "get":
            if case["initial_state"] is not None:
                manager._is_pressed = case["initial_state"]
            result = manager.get_pressed_state()
            assert result == case["expected_get"]
        else:
            manager.set_pressed_state(case["new_state"])
            assert manager._is_pressed == case["expected_set"]

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "modifier_grabbed",
                "grabbed_modifiers": {"Insert": 1, "KP_Insert": 2},
                "modifier": "Insert",
                "expected_result": True,
            },
            {
                "id": "kp_modifier_grabbed",
                "grabbed_modifiers": {"Insert": 1, "KP_Insert": 2},
                "modifier": "KP_Insert",
                "expected_result": True,
            },
            {
                "id": "modifier_not_grabbed",
                "grabbed_modifiers": {"Insert": 1},
                "modifier": "KP_Insert",
                "expected_result": False,
            },
            {
                "id": "no_grabs",
                "grabbed_modifiers": {},
                "modifier": "Insert",
                "expected_result": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_is_modifier_grabbed(
        self,
        test_context,
        case: dict,
    ) -> None:
        """Test OrcaModifierManager.is_modifier_grabbed."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        manager._grabbed_modifiers = case["grabbed_modifiers"]
        result = manager.is_modifier_grabbed(case["modifier"])
        assert result == case["expected_result"]

    @pytest.mark.parametrize(
        "orca_modifier_keys, expected_calls",
        [
            pytest.param(["Insert", "KP_Insert"], ["Insert", "KP_Insert"], id="insert_keys"),
            pytest.param(["Caps_Lock"], [], id="caps_lock_no_grab"),
            pytest.param(
                ["Insert", "Caps_Lock", "KP_Insert"], ["Insert", "KP_Insert"], id="mixed_keys"
            ),
            pytest.param([], [], id="no_keys"),
        ],
    )
    def test_add_grabs_for_orca_modifiers(
        self, test_context, orca_modifier_keys, expected_calls
    ) -> None:
        """Test OrcaModifierManager.add_grabs_for_orca_modifiers."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()

        mock_sm = test_context.Mock()
        test_context.patch(
            "orca.orca_modifier_manager.settings_manager.get_manager", new=mock_sm
        )
        mock_sm.return_value.get_setting.return_value = orca_modifier_keys
        mock_add_grab = test_context.Mock()
        test_context.patch_object(manager, "add_modifier_grab", new=mock_add_grab)
        manager.add_grabs_for_orca_modifiers()
        calls = [call(modifier) for modifier in expected_calls]
        mock_add_grab.assert_has_calls(calls, any_order=True)
        assert mock_add_grab.call_count == len(expected_calls)

    @pytest.mark.parametrize(
        "orca_modifier_keys, expected_calls",
        [
            pytest.param(["Insert", "KP_Insert"], ["Insert", "KP_Insert"], id="insert_keys"),
            pytest.param(["Caps_Lock"], [], id="caps_lock_no_ungrab"),
            pytest.param(
                ["Insert", "Caps_Lock", "KP_Insert"], ["Insert", "KP_Insert"], id="mixed_keys"
            ),
            pytest.param([], [], id="no_keys"),
        ],
    )
    def test_remove_grabs_for_orca_modifiers(
        self, test_context, orca_modifier_keys, expected_calls
    ) -> None:
        """Test OrcaModifierManager.remove_grabs_for_orca_modifiers."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()

        mock_sm = test_context.Mock()
        test_context.patch(
            "orca.orca_modifier_manager.settings_manager.get_manager", new=mock_sm
        )
        mock_sm.return_value.get_setting.return_value = orca_modifier_keys
        mock_remove_grab = test_context.Mock()
        test_context.patch_object(manager, "remove_modifier_grab", new=mock_remove_grab)
        manager.remove_grabs_for_orca_modifiers()
        calls = [call(modifier) for modifier in expected_calls]
        mock_remove_grab.assert_has_calls(calls, any_order=True)
        assert mock_remove_grab.call_count == len(expected_calls)
        assert manager._is_pressed is False

    @pytest.mark.parametrize(
        "scenario,has_existing,grab_result,expects_keycodes_call,expects_grab_call,expects_in_dict",
        [
            pytest.param("new", False, 123, True, True, True, id="new_modifier"),
            pytest.param("existing", True, None, False, False, True, id="existing_modifier"),
            pytest.param("failed", False, -1, True, True, False, id="failed_grab"),
        ],
    )
    def test_add_modifier_grab(
        self,
        test_context,
        scenario: str,
        has_existing: bool,
        grab_result: int | None,
        expects_keycodes_call: bool,
        expects_grab_call: bool,
        expects_in_dict: bool,
    ) -> None:
        """Test OrcaModifierManager.add_modifier_grab with various scenarios."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()

        if has_existing:
            manager._grabbed_modifiers["Insert"] = 123

        if scenario != "existing":
            mock_get_keycodes = test_context.Mock()
            test_context.patch(
                "orca.orca_modifier_manager.keybindings.get_keycodes", new=mock_get_keycodes
            )
            mock_iem = test_context.Mock()
            test_context.patch(
                "orca.orca_modifier_manager.input_event_manager.get_manager", new=mock_iem
            )
            mock_get_keycodes.return_value = (65379, 110)
            mock_input_manager = test_context.Mock()
            mock_input_manager.add_grab_for_modifier.return_value = grab_result
            mock_iem.return_value = mock_input_manager

        manager.add_modifier_grab("Insert")

        if expects_keycodes_call:
            mock_get_keycodes.assert_called_once_with("Insert")
        elif scenario == "existing":
            essential_modules["orca.keybindings"].get_keycodes.assert_not_called()

        if expects_grab_call:
            mock_input_manager.add_grab_for_modifier.assert_called_once_with("Insert", 65379, 110)
        elif scenario == "existing":
            essential_modules["input_manager_instance"].add_grab_for_modifier.assert_not_called()

        if expects_in_dict:
            if scenario == "new":
                assert manager._grabbed_modifiers["Insert"] == 123
            elif scenario == "existing":
                assert manager._grabbed_modifiers["Insert"] == 123
        else:
            assert "Insert" not in manager._grabbed_modifiers

    @pytest.mark.parametrize(
        "has_grabbed,expects_call",
        [
            pytest.param(True, True, id="grabbed_modifier"),
            pytest.param(False, False, id="not_grabbed_modifier"),
        ],
    )
    def test_remove_modifier_grab(
        self,
        test_context,
        has_grabbed: bool,
        expects_call: bool,
    ) -> None:
        """Test OrcaModifierManager.remove_modifier_grab with grabbed and non-grabbed modifiers."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()

        if has_grabbed:
            manager._grabbed_modifiers["Insert"] = 123
            mock_iem = test_context.Mock()
            test_context.patch(
                "orca.orca_modifier_manager.input_event_manager.get_manager", new=mock_iem
            )
            mock_input_manager = test_context.Mock()
            mock_iem.return_value = mock_input_manager

        manager.remove_modifier_grab("Insert")

        if expects_call:
            mock_input_manager.remove_grab_for_modifier.assert_called_once_with("Insert", 123)
            assert "Insert" not in manager._grabbed_modifiers
        else:
            essential_modules["input_manager_instance"].remove_grab_for_modifier.assert_not_called()

    @pytest.mark.parametrize(
        "keyval_name, expected_method",
        [
            pytest.param("Caps_Lock", "_toggle_modifier_lock", id="caps_lock"),
            pytest.param("Shift_Lock", "_toggle_modifier_lock", id="shift_lock"),
            pytest.param("Insert", "_toggle_modifier_grab", id="insert"),
            pytest.param("KP_Insert", "_toggle_modifier_grab", id="kp_insert"),
        ],
    )
    def test_toggle_modifier(
        self,
        test_context,
        keyval_name,
        expected_method,
    ) -> None:
        """Test OrcaModifierManager.toggle_modifier."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        mock_keyboard_event = test_context.Mock()
        mock_keyboard_event.keyval_name = keyval_name
        mock_keyboard_event.hw_code = 110
        mock_keyboard_event.modifiers = 0
        mock_keyboard_event.is_pressed_key.return_value = False
        mock_method = test_context.Mock()
        test_context.patch_object(manager, expected_method, new=mock_method)
        manager.toggle_modifier(mock_keyboard_event)
        mock_method.assert_called_once_with(mock_keyboard_event)

    @pytest.mark.parametrize(
        "is_pressed_key, expects_remove_call, expects_timeout_calls",
        [
            pytest.param(True, False, 0, id="pressed_key_no_action"),
            pytest.param(False, True, 2, id="released_key_full_action"),
        ],
    )
    def test_toggle_modifier_grab(
        self,
        test_context: OrcaTestContext,
        is_pressed_key,
        expects_remove_call,
        expects_timeout_calls,
    ) -> None:
        """Test OrcaModifierManager._toggle_modifier_grab with pressed and released keys."""
        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        mock_keyboard_event = test_context.Mock()
        mock_keyboard_event.keyval_name = "Insert"
        mock_keyboard_event.hw_code = 110
        mock_keyboard_event.modifiers = 0
        mock_keyboard_event.is_pressed_key.return_value = is_pressed_key
        mock_remove = test_context.Mock()
        test_context.patch_object(manager, "remove_modifier_grab", new=mock_remove)
        mock_add_grab = test_context.Mock()
        test_context.patch_object(manager, "add_modifier_grab", new=mock_add_grab)
        mock_timeout = test_context.Mock()
        test_context.patch(
            "orca.orca_modifier_manager.GLib.timeout_add", new=mock_timeout
        )
        test_context.patch(
            "orca.orca_modifier_manager.Atspi.generate_keyboard_event",
            side_effect=lambda *args, **kwargs: None
        )

        manager._toggle_modifier_grab(mock_keyboard_event)

        if expects_remove_call:
            mock_remove.assert_called_once_with("Insert")
            assert mock_timeout.call_count == expects_timeout_calls
            if expects_timeout_calls > 0:
                timeout_calls = mock_timeout.call_args_list
                assert timeout_calls[0][0][0] == 1
                assert timeout_calls[0][0][2] == 110  # hw_code
                if expects_timeout_calls > 1:
                    assert timeout_calls[1][0][0] == 500
                    assert timeout_calls[1][0][2] == "Insert"  # modifier name
        else:
            mock_remove.assert_not_called()
            assert mock_timeout.call_count == expects_timeout_calls

    @pytest.mark.parametrize(
        "keyval_name, is_pressed, expected_modifier",
        [
            pytest.param(
                "Caps_Lock",
                True,
                1 << 1,
                id="caps_lock_pressed",  # 1 << Atspi.ModifierType.SHIFTLOCK
            ),
            pytest.param(
                "Shift_Lock",
                True,
                1 << 0,
                id="shift_lock_pressed",  # 1 << Atspi.ModifierType.SHIFT
            ),
            pytest.param("Caps_Lock", False, None, id="caps_lock_released"),
            pytest.param("Other_Key", True, None, id="other_key"),
        ],
    )
    def test_toggle_modifier_lock(
        self,
        test_context,
        keyval_name,
        is_pressed,
        expected_modifier,
    ) -> None:
        """Test OrcaModifierManager._toggle_modifier_lock."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        test_context.patch(
            "orca.orca_modifier_manager.Atspi.ModifierType.SHIFTLOCK", new=1
        )
        test_context.patch("orca.orca_modifier_manager.Atspi.ModifierType.SHIFT", new=0)
        manager = orca_modifier_manager.OrcaModifierManager()
        mock_keyboard_event = test_context.Mock()
        mock_keyboard_event.keyval_name = keyval_name
        mock_keyboard_event.hw_code = 110
        mock_keyboard_event.modifiers = 0
        mock_keyboard_event.is_pressed_key.return_value = is_pressed
        mock_timeout = test_context.Mock()
        test_context.patch(
            "orca.orca_modifier_manager.GLib.timeout_add", new=mock_timeout
        )
        manager._toggle_modifier_lock(mock_keyboard_event)
        if expected_modifier is not None:
            mock_timeout.assert_called_once()
            timeout_call = mock_timeout.call_args_list[0]
            assert timeout_call[0][0] == 1  # 1ms delay
            assert timeout_call[0][2] == 0  # modifiers
            assert timeout_call[0][3] == expected_modifier  # modifier value
        else:
            mock_timeout.assert_not_called()

    def test_refresh_orca_modifiers(self, test_context: OrcaTestContext) -> None:
        """Test OrcaModifierManager.refresh_orca_modifiers."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        test_context.patch("os.environ", new={"DISPLAY": ":0"})
        mock_popen = test_context.Mock()
        test_context.patch("orca.orca_modifier_manager.subprocess.Popen", new=mock_popen)
        mock_process = test_context.Mock()
        mock_process.communicate.return_value = (b"xmodmap_content", b"")
        mock_context_manager = test_context.Mock()
        mock_context_manager.__enter__ = test_context.Mock(return_value=mock_process)
        mock_context_manager.__exit__ = test_context.Mock(return_value=None)
        mock_popen.return_value = mock_context_manager
        mock_unset = test_context.Mock()
        test_context.patch_object(manager, "unset_orca_modifiers", new=mock_unset)
        mock_create = test_context.Mock()
        test_context.patch_object(manager, "_create_orca_xmodmap", new=mock_create)
        manager.refresh_orca_modifiers("test reason")
        mock_unset.assert_called_once_with("test reason")
        mock_popen.assert_called_once_with(
            ["xkbcomp", ":0", "-"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
        )
        assert manager._original_xmodmap == b"xmodmap_content"
        mock_create.assert_called_once()

    @pytest.mark.parametrize(
        "caps_lock_orca, shift_lock_orca, caps_cleared, expected_calls",
        [
            pytest.param(True, False, False, [("set_caps", True)], id="caps_lock_enable"),
            pytest.param(False, True, False, [("set_caps", True)], id="shift_lock_enable"),
            pytest.param(True, True, False, [("set_caps", True)], id="both_enable"),
            pytest.param(
                False, False, True, [("set_caps", False)], id="disable_previously_cleared"
            ),
            pytest.param(False, False, False, [], id="no_changes"),
        ],
    )
    def test_create_orca_xmodmap(
        self,
        test_context,
        caps_lock_orca,
        shift_lock_orca,
        caps_cleared,
        expected_calls,
    ) -> None:
        """Test OrcaModifierManager._create_orca_xmodmap."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        manager._caps_lock_cleared = caps_cleared

        def is_orca_modifier_side_effect(modifier):
            if modifier == "Caps_Lock":
                return caps_lock_orca
            if modifier == "Shift_Lock":
                return shift_lock_orca
            return False

        mock_is_orca_modifier = test_context.Mock(side_effect=is_orca_modifier_side_effect)
        test_context.patch_object(manager, "is_orca_modifier", new=mock_is_orca_modifier)
        mock_set_caps = test_context.Mock()
        test_context.patch_object(manager, "set_caps_lock_as_orca_modifier", new=mock_set_caps)
        manager._create_orca_xmodmap()
        for call_type, enable in expected_calls:
            if call_type == "set_caps":
                mock_set_caps.assert_called_with(enable)

        if caps_lock_orca or shift_lock_orca:
            assert manager._caps_lock_cleared is True
        elif caps_cleared and not caps_lock_orca and not shift_lock_orca:
            assert manager._caps_lock_cleared is False

    @pytest.mark.parametrize(
        "has_xmodmap, expects_popen_call",
        [
            pytest.param(True, True, id="with_xmodmap"),
            pytest.param(False, False, id="no_xmodmap"),
        ],
    )
    def test_unset_orca_modifiers(
        self, test_context: OrcaTestContext, has_xmodmap, expects_popen_call
    ) -> None:
        """Test OrcaModifierManager.unset_orca_modifiers with and without xmodmap."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        if has_xmodmap:
            test_context.patch("os.environ", new={"DISPLAY": ":0"})
            manager._original_xmodmap = b"original_xmodmap_content"
        else:
            manager._original_xmodmap = b""

        essential_modules["orca.debug"].reset_mock()
        mock_popen = test_context.Mock()
        test_context.patch("orca.orca_modifier_manager.subprocess.Popen", new=mock_popen)

        if has_xmodmap:
            mock_process = test_context.Mock()
            mock_context_manager = test_context.Mock()
            mock_context_manager.__enter__ = test_context.Mock(return_value=mock_process)
            mock_context_manager.__exit__ = test_context.Mock(return_value=None)
            mock_popen.return_value = mock_context_manager

        manager.unset_orca_modifiers("test reason" if has_xmodmap else "")

        if expects_popen_call:
            mock_popen.assert_called_once_with(
                ["xkbcomp", "-w0", "-", ":0"], stdin=subprocess.PIPE, stdout=None, stderr=None
            )
            mock_process.communicate.assert_called_once_with(b"original_xmodmap_content")
            assert manager._caps_lock_cleared is False
        else:
            mock_popen.assert_not_called()

    @pytest.mark.parametrize(
        "enable, xmodmap_content, expects_popen_call, expected_content",
        [
            pytest.param(
                True,
                """interpret Caps_Lock+AnyOfOrNone(all) {
        action= LockMods(modifiers=Lock);
    };""",
                True,
                b"NoAction()",
                id="enable_caps_lock",
            ),
            pytest.param(
                False,
                """interpret Caps_Lock+AnyOfOrNone(all) {
        action= NoAction();
    };""",
                True,
                b"LockMods(modifiers=Lock)",
                id="disable_caps_lock",
            ),
            pytest.param(
                True,
                "some other xmodmap content",
                False,
                None,
                id="no_changes_needed",
            ),
        ],
    )
    def test_set_caps_lock_as_orca_modifier(
        self,
        test_context: OrcaTestContext,
        enable,
        xmodmap_content,
        expects_popen_call,
        expected_content,
    ) -> None:
        """Test OrcaModifierManager.set_caps_lock_as_orca_modifier with various scenarios."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager = orca_modifier_manager.OrcaModifierManager()
        if expects_popen_call:
            test_context.patch("os.environ", new={"DISPLAY": ":0"})

        manager._original_xmodmap = xmodmap_content.encode("UTF-8")
        essential_modules["orca.debug"].reset_mock()
        mock_popen = test_context.Mock()
        test_context.patch("orca.orca_modifier_manager.subprocess.Popen", new=mock_popen)

        if expects_popen_call:
            mock_process = test_context.Mock()
            mock_context_manager = test_context.Mock()
            mock_context_manager.__enter__ = test_context.Mock(return_value=mock_process)
            mock_context_manager.__exit__ = test_context.Mock(return_value=None)
            mock_popen.return_value = mock_context_manager

        manager.set_caps_lock_as_orca_modifier(enable)

        if expects_popen_call:
            mock_popen.assert_called_once_with(
                ["xkbcomp", "-w0", "-", ":0"], stdin=subprocess.PIPE, stdout=None, stderr=None
            )
            called_data = mock_process.communicate.call_args[0][0]
            assert expected_content in called_data
        else:
            mock_popen.assert_not_called()

    def test_get_manager(
        self,
        test_context,
    ) -> None:
        """Test orca_modifier_manager.get_manager."""

        self._setup_dependencies(test_context)
        from orca import orca_modifier_manager

        manager1 = orca_modifier_manager.get_manager()
        assert manager1 is not None
        assert isinstance(manager1, orca_modifier_manager.OrcaModifierManager)

        manager2 = orca_modifier_manager.get_manager()
        assert manager2 is manager1



#!/usr/bin/python
# dump-tree.py
#
# Command-line tool to dump the tree of accessible objects for a given application.
# Usage: dump-tree.py <APP_NAME>
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=invalid-name
# pylint: disable=wrong-import-position

"""Command-line tool to dump the tree of accessible objects for a given application."""

from datetime import datetime
import sys

import gi
gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

def as_string(obj: Atspi.Accessible, prefix: str) -> str:
    "Convert an accessible object to a string representation."

    timestamp = f"{datetime.now():%H:%M:%S}"
    indent = " " * (len(prefix) + len(timestamp) + 2)
    return (f"{timestamp} {prefix} "
            f"{get_role_as_string(obj)} ({hex(id(obj))}) "
            f"NAME: '{get_name(obj)}' "
            f"DESCRIPTION: '{get_description(obj)}'"
            f"\n{indent}PARENT: {get_parent_as_string(obj)} "
            f"INDEX IN PARENT: {get_index_in_parent(obj)}"
            f"\n{indent}ATTRIBUTES: {get_attributes_as_string(obj)}"
            f"\n{indent}RECT: {get_rect_as_string(obj)}"
            f"\n{indent}STATES: {get_states_as_string(obj)}"
            f"\n{indent}RELATIONS: {get_relations_as_string(obj)}"
            f"\n{indent}TEXT: {get_text(obj)}"
            )

def get_rect_as_string(obj: Atspi.Accessible) -> str:
    "Get the rectangle dimensions of an accessible object as a string."

    try:
        if Atspi.Accessible.get_role(obj) == Atspi.Role.APPLICATION:
            return ""
    except GLib.GError as error:
        print(f"Exception in get_role: {error}")
        return ""

    try:
        rect = Atspi.Component.get_extents(obj, Atspi.CoordType.WINDOW)
    except GLib.GError as error:
        print(f"Exception in get_extents: {error}")
        return ""
    return f"X:{rect.x}, Y:{rect.y}, WIDTH:{rect.width}, HEIGHT:{rect.height}"

def get_relations_as_string(obj: Atspi.Accessible) -> str:
    "Get the relations of an accessible object as a string."

    try:
        relations = Atspi.Accessible.get_relation_set(obj)
    except GLib.GError as error:
        print(f"Exception in get_relation_set: {error}")
        return ""

    result = []
    for relation in relations:
        if relation.get_n_targets() == 0:
            continue
        targets = [relation.get_target(i) for i in range(relation.get_n_targets())]
        target_addresses = ", ".join(
            [f"{get_role_as_string(target)} ({hex(id(target))})" for target in targets]
        )
        result.append(f"{relation.get_relation_type().value_nick}: {target_addresses}")

    return " ".join(result) or "(none)"

def get_states_as_string(obj: Atspi.Accessible) -> str:
    "Get the states of an accessible object as a string."

    try:
        state_set = Atspi.Accessible.get_state_set(obj)
    except GLib.GError as error:
        print(f"Exception in get_state_set: {error}")
        return ""
    return " ".join([s.value_nick for s in state_set.get_states()]) or "(none)"

def get_name(obj: Atspi.Accessible, fallback_on_labelled_by: bool = True) -> str:
    "Get the name of an accessible object."

    try:
        name = Atspi.Accessible.get_name(obj)
    except GLib.GError as error:
        print(f"Exception in get_name: {error}")
        return ""

    if name:
        return name

    result = "(none)"
    if not fallback_on_labelled_by:
        return result

    try:
        relations = Atspi.Accessible.get_relation_set(obj)
    except GLib.GError as error:
        print(f"Exception in get_relation_set: {error}")
        return ""

    for relation in relations:
        if relation.get_relation_type() != Atspi.RelationType.LABELLED_BY:
            continue
        targets = [relation.get_target(i) for i in range(relation.get_n_targets())]
        result = " ".join([get_name(target, False) for target in targets])
        if result:
            result += " (from labelled-by relation)"

    return result

def get_description(obj: Atspi.Accessible, fallback_on_described_by: bool = True) -> str:
    "Get the description of an accessible object."

    try:
        description = Atspi.Accessible.get_description(obj)
    except GLib.GError as error:
        print(f"Exception in get_description: {error}")
        return ""
    if description:
        return description

    result = "(none)"
    if not fallback_on_described_by:
        return result

    try:
        relations = Atspi.Accessible.get_relation_set(obj)
    except GLib.GError as error:
        print(f"Exception in get_relation_set: {error}")
        return ""

    for relation in relations:
        if relation.get_relation_type() != Atspi.RelationType.DESCRIBED_BY:
            continue
        targets = [relation.get_target(i) for i in range(relation.get_n_targets())]
        result = " ".join([get_name(target, False) for target in targets])
        if result:
            result += " (from described-by relation)"

    return result

def get_attributes_as_string(obj: Atspi.Accessible) -> str:
    "Get the attributes of an accessible object as a string."

    try:
        attributes = Atspi.Accessible.get_attributes(obj)
    except GLib.GError as error:
        print(f"Exception in get_attributes: {error}")
        return ""

    if not attributes:
        return "(none)"

    result = []
    for key, value in attributes.items():
        if value is None:
            value = "(null)"
        result.append(f"{key}={value}")

    return ", ".join(result)

def get_role_as_string(obj: Atspi.Accessible) -> str:
    "Get the role of an accessible object."

    try:
        return Atspi.Accessible.get_role(obj).value_name
    except GLib.GError as error:
        print(f"Exception in get_role: {error}")
        return "(unknown role)"

def get_text(obj: Atspi.Accessible) -> str:
    "Get the text of an accessible object."

    try:
        length = Atspi.Text.get_character_count(obj)
        result = Atspi.Text.get_text(obj, 0, length)
    except GLib.GError:
        return "(not implemented)"

    result = result.replace("\n", "\\n")
    if len(result) > 80:
        result = result[:80] + "[...]"
    return f"'{result}'"

def get_parent(obj: Atspi.Accessible) -> Atspi.Accessible:
    "Get the parent of an accessible object."

    try:
        if Atspi.Accessible.get_role(obj) == Atspi.Role.APPLICATION:
            return None
    except GLib.GError as error:
        print(f"Exception in get_role: {error}")
        return None

    try:
        return Atspi.Accessible.get_parent(obj)
    except GLib.GError as error:
        print(f"Exception in get_parent: {error}")
        return None

def get_parent_as_string(obj: Atspi.Accessible) -> str:
    "Get the parent of an accessible object."

    parent = get_parent(obj)
    if parent:
        return f"{get_role_as_string(parent)} ({hex(id(parent))})"
    return ""

def get_index_in_parent(obj: Atspi.Accessible) -> int:
    "Get the index of an accessible object in its parent's children."

    parent = get_parent(obj)
    if not parent:
        return -1

    try:
        index = Atspi.Accessible.get_index_in_parent(obj)
    except GLib.GError as error:
        print(f"Exception in get_index_in_parent: {error}")
        return -1

    if index < 0:
        print(f"WARNING: Object {hex(id(obj))} has invalid index in parent {hex(id(parent))}")
        return -1

    child = Atspi.Accessible.get_child_at_index(parent, index)
    if child != obj:
        print(f"WARNING: Object {hex(id(obj))} at index {index} in parent {hex(id(parent))} "
              f"is not the expected child {hex(id(child))}")

    return index

def get_children(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
    "Get the children of an accessible object."

    children = []
    try:
        count = Atspi.Accessible.get_child_count(obj)
    except GLib.GError as error:
        print(f"Exception in get_child_count: {error}")
        return []

    for i in range(count):
        try:
            children.append(Atspi.Accessible.get_child_at_index(obj, i))
        except GLib.GError as error:
            print(f"Exception in get_child_at_index: {error}")
            return children

    return children

def clear_cache_single(obj: Atspi.Accessible) -> None:
    "Clear the cache for a single accessible object."

    try:
        Atspi.Accessible.clear_cache_single(obj)
    except GLib.GError as error:
        print(f"Exception in clear_cache_single: {error}")

def print_tree(root: Atspi.Accessible, indent: int = 0) -> None:
    "Print the tree structure of accessible objects."

    clear_cache_single(root)
    prefix = f"{'  ' * indent}-->"
    print(f"{as_string(root, prefix)}")
    for child in get_children(root):
        print_tree(child, indent + 1)

def find_application_with_name(app_name: str) -> Atspi.Accessible:
    "Find the accessible application with the specified name."

    try:
        desktop = Atspi.get_desktop(0)
    except GLib.GError as error:
        print(f"Exception getting desktop from Atspi: {error}")
        return None

    apps = []
    for child in get_children(desktop):
        if get_name(child, False) == app_name:
            return child
        apps.append(child)

    apps_as_string = "\n".join([f"{i + 1}. {get_name(app, False)}" for i, app in enumerate(apps)])
    print(
        f"Application '{app_name}' not found.\nDesktop has {len(apps)} accessible applications:\n"
        f"{apps_as_string}"
    )
    return None

def main():
    """Starts the tree dumper and waits for events."""

    if len(sys.argv) != 2:
        print("Usage: dump-tree.py <APP_NAME>")
        sys.exit(1)

    app = find_application_with_name(sys.argv[1])
    if not app:
        sys.exit(1)

    print_tree(app)
    print("Tree dump complete.")

if __name__ == "__main__":
    main()



# Orca
#
# Copyright (C) 2023 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Custom script utilities for Qt"""

# This has to be the first non-docstring line in the module to make linters happy.
from __future__ import annotations

__id__ = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2023 Igalia, S.L."
__license__   = "LGPL"

from typing import TYPE_CHECKING

from orca import debug
from orca import script_utilities
from orca.ax_object import AXObject
from orca.ax_utilities import AXUtilities

if TYPE_CHECKING:
    import gi
    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

class Utilities(script_utilities.Utilities):
    """Custom script utilities for Qt"""

    def _is_top_level_object(self, obj: Atspi.Accessible) -> bool:
        # This is needed because Qt apps might insert some junk (e.g. a filler) in
        # between the window/frame/dialog and the application.
        return AXUtilities.is_application(AXObject.get_parent(obj))

    def top_level_object(
        self,
        obj: Atspi.Accessible,
        use_fallback_search: bool = False
    ) -> Atspi.Accessible | None:
        """Returns the top level object for obj."""

        # The fallback search is needed because sometimes we can ascend the accessibility
        # tree all the way to the top; other times, we cannot get the parent, but can still
        # get the children. The fallback search handles the latter scenario.
        result = super().top_level_object(obj, use_fallback_search=True)
        if result is not None and AXObject.get_role(result) not in self._top_level_roles():
            tokens = ["QT: Top level object", result, "lacks expected role."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    def frame_and_dialog(
        self,
        obj: Atspi.Accessible | None = None
    ) -> list[Atspi.Accessible | None]:
        """Returns the frame and (possibly) the dialog containing obj."""

        frame, dialog = super().frame_and_dialog(obj)
        if frame or dialog:
            return [frame, dialog]

        # https://bugreports.qt.io/browse/QTBUG-129656
        tokens = ["QT: Could not find frame or dialog for", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        top_level = self.top_level_object(obj, True)

        tokens = ["QT: Returning", top_level, "as frame for", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return [top_level, None]

    def has_meaningful_toggle_action(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj has a meaningful toggle action."""

        # https://bugreports.qt.io/browse/QTBUG-116204
        if AXUtilities.is_table_cell_or_header(obj):
            tokens = ["QT: Ignoring toggle action on", obj, "."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        return super().has_meaningful_toggle_action(obj)



# Unit tests for sleep_mode_manager.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=protected-access
# pylint: disable=too-many-locals

"""Unit tests for sleep_mode_manager.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import gi
import pytest

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext
    from unittest.mock import MagicMock

@pytest.mark.unit
class TestSleepModeManager:
    """Test SleepModeManager class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Returns dependencies for sleep_mode_manager module testing."""

        essential_modules = test_context.setup_shared_dependencies([])

        debug_mock = essential_modules["orca.debug"]
        debug_mock.print_message = test_context.Mock()
        debug_mock.print_tokens = test_context.Mock()
        debug_mock.debugFile = None
        debug_mock.LEVEL_INFO = 800

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_mock.get_name = test_context.Mock(return_value="TestApp")

        braille_mock = essential_modules["orca.braille"]
        braille_mock.clear = test_context.Mock()

        cmdnames_mock = essential_modules["orca.cmdnames"]
        cmdnames_mock.TOGGLE_SLEEP_MODE = "Toggle sleep mode for the current application"

        dbus_service_mock = essential_modules["orca.dbus_service"]
        controller_mock = test_context.Mock()
        controller_mock.register_decorated_module = test_context.Mock()
        dbus_service_mock.get_remote_controller = test_context.Mock(
            return_value=controller_mock
        )
        dbus_service_mock.command = lambda func: func

        input_event_mock = essential_modules["orca.input_event"]
        input_event_handler_mock = test_context.Mock()
        input_event_mock.InputEventHandler = test_context.Mock(
            return_value=input_event_handler_mock
        )

        keybindings_mock = essential_modules["orca.keybindings"]
        key_bindings_instance = test_context.Mock()
        key_bindings_instance.is_empty = test_context.Mock(return_value=False)
        key_bindings_instance.remove_key_grabs = test_context.Mock(return_value=None)
        key_bindings_instance.add = test_context.Mock(return_value=None)
        keybindings_mock.KeyBindings = test_context.Mock(return_value=key_bindings_instance)
        keybindings_mock.KeyBinding = test_context.Mock()
        keybindings_mock.DEFAULT_MODIFIER_MASK = 1
        keybindings_mock.SHIFT_ALT_CTRL_MODIFIER_MASK = 15

        messages_mock = essential_modules["orca.messages"]
        messages_mock.SLEEP_MODE_ENABLED_FOR = "Sleep mode enabled for %s."
        messages_mock.SLEEP_MODE_DISABLED_FOR = "Sleep mode disabled for %s."

        script_manager_mock = essential_modules["orca.script_manager"]
        script_manager_instance = test_context.Mock()
        script_manager_instance.get_or_create_sleep_mode_script = test_context.Mock()
        script_manager_instance.set_active_script = test_context.Mock()
        script_manager_instance.get_script = test_context.Mock()
        script_manager_mock.get_manager = test_context.Mock(
            return_value=script_manager_instance
        )

        essential_modules["controller"] = controller_mock
        essential_modules["input_event_handler"] = input_event_handler_mock
        essential_modules["key_bindings_instance"] = key_bindings_instance
        essential_modules["script_manager_instance"] = script_manager_instance

        return essential_modules

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test SleepModeManager.__init__ with default parameters."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        assert manager._handlers is not None
        assert manager._bindings is not None
        assert not manager._apps
        essential_modules["orca.keybindings"].KeyBindings.assert_called()

        assert essential_modules["orca.dbus_service"] is not None
        assert essential_modules["controller"] is not None

    @pytest.mark.parametrize(
        "is_desktop",
        [
            pytest.param(True, id="desktop"),
            pytest.param(False, id="not_desktop"),
        ],
    )
    def test_get_bindings_refresh(self, test_context: OrcaTestContext, is_desktop: bool) -> None:
        """Test SleepModeManager.get_bindings with refresh=True and various desktop settings."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        bindings = manager.get_bindings(refresh=True, is_desktop=is_desktop)
        assert bindings is not None
        essential_modules["orca.debug"].print_message.assert_any_call(
            essential_modules["orca.debug"].LEVEL_INFO,
            f"SLEEP MODE MANAGER: Refreshing bindings. Is desktop: {is_desktop}",
            True,
        )

    def test_get_handlers_refresh_true(self, test_context: OrcaTestContext) -> None:
        """Test SleepModeManager.get_handlers with refresh=True."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        handlers = manager.get_handlers(refresh=True)
        assert handlers is not None
        essential_modules["orca.debug"].print_message.assert_any_call(
            essential_modules["orca.debug"].LEVEL_INFO,
            "SLEEP MODE MANAGER: Refreshing handlers.",
            True,
        )

    def test_is_active_for_app_with_active_app(self, test_context: OrcaTestContext) -> None:
        """Test SleepModeManager.is_active_for_app returns True for active app."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        mock_app = test_context.Mock(spec=Atspi.Accessible)
        app_hash = hash(mock_app)
        manager._apps.append(app_hash)

        essential_modules["orca.debug"].print_tokens.reset_mock()
        result = manager.is_active_for_app(mock_app)
        assert result is True
        essential_modules["orca.debug"].print_tokens.assert_called_with(
            essential_modules["orca.debug"].LEVEL_INFO,
            ["SLEEP MODE MANAGER: Is active for", mock_app],
            True,
        )

    def test_setup_handlers(self, test_context: OrcaTestContext) -> None:
        """Test SleepModeManager._setup_handlers method."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        manager._setup_handlers()
        assert "toggle_sleep_mode" in manager._handlers
        expected_handler = essential_modules["input_event_handler"]
        assert manager._handlers["toggle_sleep_mode"] == expected_handler
        input_event = essential_modules["orca.input_event"]
        input_event.InputEventHandler.assert_called_with(
            manager.toggle_sleep_mode, "Toggle sleep mode for the current application"
        )

    def test_setup_bindings(self, test_context: OrcaTestContext) -> None:
        """Test SleepModeManager._setup_bindings method."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        manager._setup_bindings()

        essential_modules["orca.keybindings"].KeyBinding.assert_called_with(
            "q",
            1,  # DEFAULT_MODIFIER_MASK
            15,  # SHIFT_ALT_CTRL_MODIFIER_MASK
            manager._handlers["toggle_sleep_mode"],
        )

        key_bindings_instance = essential_modules["key_bindings_instance"]
        key_bindings_instance.add.assert_called()

    def test_toggle_sleep_mode_no_script(self, test_context: OrcaTestContext) -> None:
        """Test SleepModeManager.toggle_sleep_mode with no script."""

        self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        result = manager.toggle_sleep_mode(None)
        assert result is True

    def test_toggle_sleep_mode_script_no_app(self, test_context: OrcaTestContext) -> None:
        """Test SleepModeManager.toggle_sleep_mode with script but no app."""

        self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        mock_script = test_context.Mock()
        mock_script.app = None
        result = manager.toggle_sleep_mode(mock_script)
        assert result is True

    @pytest.mark.parametrize(
        "initially_active,notify_user,expected_message_contains",
        [
            pytest.param(False, True, "Sleep mode enabled for", id="enable_with_notification"),
            pytest.param(False, False, None, id="enable_without_notification"),
            pytest.param(True, True, "Sleep mode disabled for", id="disable_with_notification"),
            pytest.param(True, False, None, id="disable_without_notification"),
        ],
    )
    def test_toggle_sleep_mode_with_script_and_app(  # pylint: disable=too-many-branches
        self,
        test_context: OrcaTestContext,
        initially_active: bool,
        notify_user: bool,
        expected_message_contains: str | None,
    ) -> None:
        """Test SleepModeManager.toggle_sleep_mode with various scenarios."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.sleep_mode_manager import SleepModeManager

        manager = SleepModeManager()
        mock_script = test_context.Mock()
        mock_app = test_context.Mock()
        mock_script.app = mock_app
        mock_event = test_context.Mock()
        app_hash = hash(mock_app)

        if initially_active:
            manager._apps.append(app_hash)

        script_manager_instance = essential_modules["script_manager_instance"]

        if initially_active:
            new_script = test_context.Mock()
            script_manager_instance.get_script = test_context.Mock(return_value=new_script)
        else:
            sleep_script = test_context.Mock()
            script_manager_instance.get_or_create_sleep_mode_script = test_context.Mock(
                return_value=sleep_script
            )

        test_context.patch(
            "orca.sleep_mode_manager.script_manager.get_manager",
            return_value=script_manager_instance
        )

        result = manager.toggle_sleep_mode(mock_script, mock_event, notify_user=notify_user)
        assert result is True

        if initially_active:
            assert app_hash not in manager._apps
        else:
            assert app_hash in manager._apps

        if not initially_active:
            essential_modules["orca.braille"].clear.assert_called()

        if notify_user and expected_message_contains:
            if initially_active:
                new_script.present_message.assert_called_once()
                call_args = new_script.present_message.call_args[0][0]
            else:
                mock_script.present_message.assert_called_once()
                call_args = mock_script.present_message.call_args[0][0]
            assert expected_message_contains in call_args
        else:
            if initially_active:
                new_script.present_message.assert_not_called()
            else:
                mock_script.present_message.assert_not_called()

        if initially_active:
            script_manager_instance.get_script.assert_called_with(mock_app)
            script_manager_instance.set_active_script.assert_called_with(
                new_script, "Sleep mode toggled off"
            )
        else:
            script_manager_instance.get_or_create_sleep_mode_script.assert_called_with(mock_app)
            script_manager_instance.set_active_script.assert_called_with(
                sleep_script, "Sleep mode toggled on"
            )



# Utilities for performing tasks related to accessibility inspection.
#
# Copyright 2023-2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-branches
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=wrong-import-position

"""Utilities for performing tasks related to accessibility inspection."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2023-2025 Igalia, S.L."
__license__   = "LGPL"

import functools
import inspect
import queue
import threading
import time

import gi
gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import debug
from .ax_component import AXComponent
from .ax_object import AXObject
from .ax_selection import AXSelection
from .ax_table import AXTable
from .ax_text import AXText
from .ax_utilities_application import AXUtilitiesApplication
from .ax_utilities_collection import AXUtilitiesCollection
from .ax_utilities_event import AXUtilitiesEvent
from .ax_utilities_relation import AXUtilitiesRelation
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_state import AXUtilitiesState


class AXUtilities:
    """Utilities for performing tasks related to accessibility inspection."""

    COMPARE_COLLECTION_PERFORMANCE = False

    # Things we cache.
    SET_MEMBERS: dict[int, list[Atspi.Accessible]] = {}
    IS_LAYOUT_ONLY: dict[int, tuple[bool, str]] = {}

    _lock = threading.Lock()

    @staticmethod
    def start_cache_clearing_thread() -> None:
        """Starts thread to periodically clear cached details."""

        thread = threading.Thread(target=AXUtilities._clear_stored_data)
        thread.daemon = True
        thread.start()

    @staticmethod
    def _clear_stored_data() -> None:
        """Clears any data we have cached for objects"""

        while True:
            time.sleep(60)
            AXUtilities._clear_all_dictionaries()

    @staticmethod
    def _clear_all_dictionaries(reason: str = "") -> None:
        msg = "AXUtilities: Clearing cache."
        if reason:
            msg += f" Reason: {reason}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        with AXUtilities._lock:
            AXUtilities.SET_MEMBERS.clear()
            AXUtilities.IS_LAYOUT_ONLY.clear()

    @staticmethod
    def clear_all_cache_now(obj: Atspi.Accessible | None = None, reason: str = "") -> None:
        """Clears all cached information immediately."""

        AXUtilities._clear_all_dictionaries(reason)
        AXObject.clear_cache_now(reason)
        AXUtilitiesRelation.clear_cache_now(reason)
        AXUtilitiesEvent.clear_cache_now(reason)
        if AXUtilitiesRole.is_table_related(obj):
            AXTable.clear_cache_now(reason)

    @staticmethod
    def can_be_active_window(window: Atspi.Accessible) -> bool:
        """Returns True if window can be the active window based on its state."""

        if window is None:
            return False

        AXObject.clear_cache(window, False, "Checking if window can be the active window")
        app = AXUtilitiesApplication.get_application(window)
        tokens = ["AXUtilities:", window, "from", app]

        if not AXUtilitiesState.is_active(window):
            tokens.append("lacks active state")
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if not AXUtilitiesState.is_showing(window):
            tokens.append("lacks showing state")
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXUtilitiesState.is_iconified(window):
            tokens.append("is iconified")
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXObject.get_name(app) == "mutter-x11-frames":
            tokens.append("is from app that cannot have the real active window")
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if app and not AXUtilitiesApplication.is_application_in_desktop(app):
            tokens.append("is from app unknown to AT-SPI2")
            # Firefox alerts and dialogs suffer from this bug too, but if we ignore these windows
            # we'll fail to fully present things like the file chooser dialog and the replace-file
            # alert. https://bugzilla.mozilla.org/show_bug.cgi?id=1882794
            if not AXUtilitiesRole.is_dialog_or_alert(window):
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        tokens.append("can be active window")
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    @staticmethod
    def find_active_window() -> Atspi.Accessible | None:
        """Tries to locate the active window; may or may not succeed."""

        candidates = []
        apps = AXUtilitiesApplication.get_all_applications(must_have_window=True)
        for app in apps:
            candidates.extend(list(AXObject.iter_children(app, AXUtilities.can_be_active_window)))

        if not candidates:
            tokens = ["AXUtilities: Unable to find active window from", apps]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if len(candidates) == 1:
            tokens = ["AXUtilities: Active window is", candidates[0]]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return candidates[0]

        tokens = ["AXUtilities: These windows all claim to be active:", candidates]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # Some electron apps running in the background claim to be active even when they
        # are not. These are the ones we know about. We can add others as we go.
        suspect_apps = ["slack",
                        "discord",
                        "outline-client",
                        "whatsapp-desktop-linux"]
        filtered = []
        for frame in candidates:
            if AXObject.get_name(AXUtilitiesApplication.get_application(frame)) in suspect_apps:
                tokens = ["AXUtilities: Suspecting", frame, "is a non-active Electron app"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            else:
                filtered.append(frame)

        if len(filtered) == 1:
            tokens = ["AXUtilities: Active window is believed to be", filtered[0]]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return filtered[0]

        guess: Atspi.Accessible | None = None
        if filtered:
            tokens = ["AXUtilities: Still have multiple active windows:", filtered]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            guess = filtered[0]

        if guess is not None:
            tokens = ["AXUtilities: Returning", guess, "as active window"]
        else:
            tokens = ["AXUtilities: No active window found"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return guess

    @staticmethod
    def is_unfocused_alert_or_dialog(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is an unfocused alert or dialog with presentable items."""

        if not AXUtilitiesRole.is_dialog_or_alert(obj):
            return False
        if not AXObject.get_child_count(obj):
            return False
        if not AXUtilitiesState.is_showing(obj):
            return False
        return not AXUtilities.can_be_active_window(obj)

    @staticmethod
    def get_unfocused_alerts_and_dialogs(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns a list of all the unfocused alerts and dialogs in the app and window of obj."""

        app = AXUtilitiesApplication.get_application(obj)
        result = list(AXObject.iter_children(app, AXUtilities.is_unfocused_alert_or_dialog))

        frame = AXObject.find_ancestor(
            obj, lambda x: AXUtilitiesRole.is_application(AXObject.get_parent(x)))
        result.extend(list(AXObject.iter_children(frame, AXUtilities.is_unfocused_alert_or_dialog)))

        tokens = ["AXUtilities: Unfocused alerts and dialogs for", obj, ":", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_all_widgets(
        obj: Atspi.Accessible,
        must_be_showing_and_visible: bool = True,
        exclude_push_button: bool = False
    ) -> list[Atspi.Accessible]:
        """Returns all the descendants of obj with a widget role"""

        roles = AXUtilitiesRole.get_widget_roles()
        if exclude_push_button and Atspi.Role.PUSH_BUTTON in roles:
            roles.remove(Atspi.Role.PUSH_BUTTON)

        result = None
        if AXObject.supports_collection(obj):
            if not must_be_showing_and_visible:
                result = AXUtilitiesCollection.find_all_with_role(obj, roles)
            else:
                states = [Atspi.StateType.SHOWING, Atspi.StateType.VISIBLE]
                result = AXUtilitiesCollection.find_all_with_role_and_all_states(
                    obj, roles, states)

            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        def is_match(acc):
            if AXObject.get_role(acc) not in roles:
                return False
            if must_be_showing_and_visible:
                return AXUtilitiesState.is_showing(acc) and AXUtilitiesState.is_visible(acc)
            return True

        return AXObject.find_all_descendants(obj, is_match)

    @staticmethod
    def get_default_button(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the default button descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_default_button(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXObject.find_descendant(obj, AXUtilitiesRole.is_default_button)

    @staticmethod
    def get_focused_object(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the focused descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_focused_object(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXObject.find_descendant(obj, AXUtilitiesState.is_focused)

    @staticmethod
    def get_info_bar(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the info bar descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_info_bar(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXObject.find_descendant(obj, AXUtilitiesRole.is_info_bar)

    @staticmethod
    def get_status_bar(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the status bar descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_status_bar(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXObject.find_descendant(obj, AXUtilitiesRole.is_status_bar)

    @staticmethod
    def _is_layout_only(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True and a string reason if obj is believed to serve only for layout."""

        reason = ""
        role = AXObject.get_role(obj)
        if role in AXUtilitiesRole.get_layout_only_roles():
            return True, "has layout-only role"

        if AXUtilitiesRole.is_layered_pane(obj, role):
            result = AXObject.find_ancestor(obj, AXUtilitiesRole.is_desktop_frame) is not None
            if result:
                reason = "is inside desktop frame"
            return result, reason

        if AXUtilitiesRole.is_menu(obj, role) or AXUtilitiesRole.is_list(obj, role):
            result = AXObject.find_ancestor(obj, AXUtilitiesRole.is_combo_box) is not None
            if result:
                reason = "is inside combo box"
            return result, reason

        if AXUtilitiesRole.is_group(obj, role):
            result = not AXUtilities.has_explicit_name(obj)
            if result:
                reason = "lacks explicit name"
            return result, reason

        if AXUtilitiesRole.is_panel(obj, role) or AXUtilitiesRole.is_grouping(obj, role):
            name = AXObject.get_name(obj)
            description = AXObject.get_description(obj)
            labelled_by = AXUtilitiesRelation.get_is_labelled_by(obj)
            described_by = AXUtilitiesRelation.get_is_described_by(obj)
            if not (name or description or labelled_by or described_by):
                return True, "lacks name, description, and relations"
            if name == AXObject.get_name(AXUtilitiesApplication.get_application(obj)):
                return True, "has same name as app"
            if AXObject.get_child_count(obj) == 1:
                child = AXObject.get_child(obj, 0)
                if name == AXObject.get_name(child):
                    return True, "has same name as its only child"
                if not AXUtilitiesRole.is_label(child) and child in labelled_by:
                    return True, "is labelled by non-label only child"
            set_roles = AXUtilitiesRole.get_set_container_roles()
            ancestor = AXObject.find_ancestor(obj, lambda x: AXObject.get_role(x) in set_roles)
            if ancestor and AXObject.get_name(ancestor) == name:
                return True, "is in set container with same name"
            return False, reason

        if AXUtilitiesRole.is_section(obj, role) or AXUtilitiesRole.is_document(obj, role):
            if AXUtilitiesState.is_focusable(obj):
                return False, "is focusable"
            if AXObject.has_action(obj, "click"):
                return False, "has click action"
            return True, "is not interactive"

        if AXUtilitiesRole.is_tool_bar(obj):
            result = AXUtilitiesRole.is_page_tab_list(AXObject.get_child(obj, 0))
            if result:
                reason = "is parent of page tab list"
            return result, reason

        if AXUtilitiesRole.is_table(obj, role):
            result = AXTable.is_layout_table(obj)
            if result:
                reason = "is layout table"
            return result, reason

        if AXUtilitiesRole.is_table_row(obj):
            if AXUtilitiesState.is_focusable(obj):
                return False, "is focusable"
            if AXUtilitiesState.is_selectable(obj):
                return False, "is selectable"
            if AXUtilitiesState.is_expandable(obj):
                return False, "is expandable"
            if AXUtilities.has_explicit_name(obj):
                return False, "has explicit name"
            return True, "is not focusable, selectable, or expandable and lacks explicit name"

        if AXUtilitiesRole.is_table_cell(obj, role):
            if AXUtilitiesRole.is_table_cell(AXObject.get_child(obj, 0)):
                return True, "child of this cell is table cell"
            table = AXTable.get_table(obj)
            if AXUtilitiesRole.is_table(table):
                result = AXTable.is_layout_table(table)
                if result:
                    reason = "is in layout table"
                return result, reason

        return False, reason

    @staticmethod
    def is_layout_only(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is believed to serve only for layout."""

        if hash(obj) in AXUtilities.IS_LAYOUT_ONLY:
            result, reason = AXUtilities.IS_LAYOUT_ONLY.get(hash(obj), (False, ""))
        else:
            result, reason = AXUtilities._is_layout_only(obj)
            AXUtilities.IS_LAYOUT_ONLY[hash(obj)] = result, reason

        if reason:
            tokens = ["AXUtilities:", obj, f"believed to be layout only: {result}, {reason}"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    @staticmethod
    def is_message_dialog(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a dialog that should be treated as a message dialog"""

        if not AXUtilitiesRole.is_dialog_or_alert(obj):
            return False

        if not AXObject.supports_collection(obj):
            widgets = AXUtilities.get_all_widgets(obj, exclude_push_button=True)
            return not widgets

        if AXUtilitiesCollection.has_scroll_pane(obj):
            tokens = ["AXUtilities:", obj, "is not a message dialog: has scroll pane"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXUtilitiesCollection.has_split_pane(obj):
            tokens = ["AXUtilities:", obj, "is not a message dialog: has split pane"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXUtilitiesCollection.has_tree_or_tree_table(obj):
            tokens = ["AXUtilities:", obj, "is not a message dialog: has tree or tree table"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXUtilitiesCollection.has_combo_box_or_list_box(obj):
            tokens = ["AXUtilities:", obj, "is not a message dialog: has combo box or list box"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXUtilitiesCollection.has_editable_object(obj):
            tokens = ["AXUtilities:", obj, "is not a message dialog: has editable object"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXUtilities:", obj, "is believed to be a message dialog"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    @staticmethod
    def is_redundant_object(obj1: Atspi.Accessible, obj2: Atspi.Accessible) -> bool:
        """Returns True if obj2 is redundant to obj1."""

        if obj1 == obj2:
            return False

        if AXObject.get_name(obj1) != AXObject.get_name(obj2) \
           or AXObject.get_role(obj1) != AXObject.get_role(obj2):
            return False

        tokens = ["AXUtilities:", obj2, "is redundant to", obj1]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    @staticmethod
    def _sort_by_child_index(object_list: list[Atspi.Accessible]) -> list[Atspi.Accessible]:
        """Returns the list of objects sorted according to child index."""

        def cmp(x, y):
            return AXObject.get_index_in_parent(x) - AXObject.get_index_in_parent(y)

        if not object_list or len(object_list) == 1:
            return object_list

        result = sorted(object_list, key=functools.cmp_to_key(cmp))

        first, second = result[0:2]
        if AXUtilitiesRole.is_radio_button(first) and AXObject.get_toolkit_name(first) == "gtk":
            # Gtk radio buttons are often in reverse order, except for when they're not.
            # See https://gitlab.gnome.org/GNOME/gtk/-/issues/7839.
            sorted_first, _sorted_second = AXComponent.sort_objects_by_position([first, second])
            if sorted_first != first:
                result.reverse()

        if object_list != result:
            tokens = ["AXUtilities: Original list", object_list]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            tokens = ["AXUtilities: Sorted list", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    @staticmethod
    def _get_set_members(
        obj: Atspi.Accessible, container: Atspi.Accessible
    ) -> list[Atspi.Accessible]:
        """Returns the members of the container of obj"""

        if container is None:
            tokens = ["AXUtilities: Members of", obj, "not obtainable: container is None"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return []

        result = AXUtilitiesRelation.get_is_member_of(obj)
        if result:
            tokens = ["AXUtilities: Members of", obj, "in", container, "via member-of", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return AXUtilities._sort_by_child_index(result)

        result = AXUtilitiesRelation.get_is_node_parent_of(obj)
        if result:
            tokens = ["AXUtilities: Members of", obj, "in", container, "via node-parent-of", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return AXUtilities._sort_by_child_index(result)

        if AXUtilitiesRole.is_description_value(obj):
            previous_sibling = AXObject.get_previous_sibling(obj)
            while previous_sibling and AXUtilitiesRole.is_description_value(previous_sibling):
                result.append(previous_sibling)
                previous_sibling = AXObject.get_previous_sibling(previous_sibling)
            result.append(obj)
            next_sibling = AXObject.get_next_sibling(obj)
            while next_sibling and AXUtilitiesRole.is_description_value(next_sibling):
                result.append(next_sibling)
                next_sibling = AXObject.get_next_sibling(next_sibling)
            tokens = ["AXUtilities: Members of", obj, "in", container, "based on siblings", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return result

        if AXUtilitiesRole.is_menu_related(obj):
            result = list(AXObject.iter_children(container, AXUtilitiesRole.is_menu_related))
            tokens = ["AXUtilities: Members of", obj, "in", container, "based on menu role", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return result

        role = AXObject.get_role(obj)
        result = list(AXObject.iter_children(container, lambda x: AXObject.get_role(x) == role))
        tokens = ["AXUtilities: Members of", obj, "in", container, "based on role", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_set_members(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns the members of the container of obj."""

        result: list[Atspi.Accessible] = []
        container = AXObject.get_parent_checked(obj)
        if hash(container) in AXUtilities.SET_MEMBERS:
            result = AXUtilities.SET_MEMBERS.get(hash(container), [])

        if obj not in result:
            if result:
                tokens = ["AXUtilities:", obj, "not in cached members of", container, ":", result]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            result = AXUtilities._get_set_members(obj, container)
            AXUtilities.SET_MEMBERS[hash(container)] = result

        # In a collapsed combobox, one can arrow to change the selection without showing the items.
        must_be_showing = not AXObject.find_ancestor(obj, AXUtilitiesRole.is_combo_box)
        if not must_be_showing:
            return result

        filtered = list(filter(AXUtilitiesState.is_showing, result))
        if result != filtered:
            tokens = ["AXUtilities: Filtered non-showing:", set(result).difference(set(filtered))]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return filtered

    @staticmethod
    def get_set_size(obj: Atspi.Accessible) -> int:
        """Returns the total number of objects in this container."""

        result = AXObject.get_attribute(obj, "setsize", False)
        if isinstance(result, str) and result.isnumeric():
            return int(result)

        if AXUtilitiesRole.is_table_row(obj):
            return AXTable.get_row_count(AXTable.get_table(obj))

        if AXUtilitiesRole.is_table_cell_or_header(obj) \
           and not AXUtilitiesRole.is_table_row(AXObject.get_parent(obj)):
            return AXTable.get_row_count(AXTable.get_table(obj))

        if AXUtilitiesRole.is_combo_box(obj):
            selected_children = AXSelection.get_selected_children(obj)
            if not selected_children:
                return -1
            if len(selected_children) == 1:
                obj = selected_children[0]

        if AXUtilitiesRole.is_list(obj) or AXUtilitiesRole.is_list_box(obj):
            obj = AXObject.find_descendant(obj, AXUtilitiesRole.is_list_item)

        child_count = AXObject.get_child_count(AXObject.get_parent(obj))
        if child_count > 500:
            return child_count

        members = AXUtilities.get_set_members(obj)
        return len(members)

    @staticmethod
    def get_set_size_is_unknown(obj: Atspi.Accessible) -> bool:
        """Returns True if the total number of objects in this container is unknown."""

        if AXUtilitiesState.is_indeterminate(obj):
            return True

        attrs = AXObject.get_attributes_dict(obj, False)
        if attrs.get("setsize") == "-1":
            return True

        if AXUtilitiesRole.is_table(obj):
            return attrs.get("rowcount") == "-1" or attrs.get("colcount") == "-1"

        return False

    @staticmethod
    def get_position_in_set(obj: Atspi.Accessible) -> int:
        """Returns the position of obj with respect to the number of items in its container."""

        result = AXObject.get_attribute(obj, "posinset", False)
        if isinstance(result, str) and result.isnumeric():
            # ARIA posinset is 1-based.
            return int(result) - 1

        if AXUtilitiesRole.is_table_row(obj):
            result = AXObject.get_attribute(obj, "rowindex", False)
            if isinstance(result, str) and result.isnumeric():
                # ARIA posinset is 1-based.
                return int(result) - 1

            if AXObject.get_child_count(obj):
                cell = AXObject.find_descendant(obj, AXUtilitiesRole.is_table_cell_or_header)
                result = AXObject.get_attribute(cell, "rowindex", False)

            if isinstance(result, str) and result.isnumeric():
                # ARIA posinset is 1-based.
                return int(result) - 1

        if AXUtilitiesRole.is_table_cell_or_header(obj) \
           and not AXUtilitiesRole.is_table_row(AXObject.get_parent(obj)):
            return AXTable.get_cell_coordinates(obj)[0]

        if AXUtilitiesRole.is_combo_box(obj):
            selected_children = AXSelection.get_selected_children(obj)
            if not selected_children:
                return -1
            if len(selected_children) == 1:
                obj = selected_children[0]

        child_count = AXObject.get_child_count(AXObject.get_parent(obj))
        if child_count > 500:
            return AXObject.get_index_in_parent(obj)

        members = AXUtilities.get_set_members(obj)
        if obj not in members:
            return -1

        return members.index(obj)

    @staticmethod
    def has_explicit_name(obj: Atspi.Accessible) -> bool:
        """Returns True if obj has an author/app-provided name as opposed to a calculated name."""

        return AXObject.get_attribute(obj, "explicit-name") == "true"

    @staticmethod
    def has_visible_caption(obj: Atspi.Accessible) -> bool:
        """Returns True if obj has a visible caption."""

        if not (AXUtilitiesRole.is_figure(obj) or AXObject.supports_table(obj)):
            return False

        labels = AXUtilitiesRelation.get_is_labelled_by(obj)
        for label in labels:
            if AXUtilitiesRole.is_caption(label) \
               and AXUtilitiesState.is_showing(label) and AXUtilitiesState.is_visible(label):
                return True

        return False

    @staticmethod
    def get_displayed_label(obj: Atspi.Accessible) -> str:
        """Returns the displayed label of obj."""

        labels = AXUtilitiesRelation.get_is_labelled_by(obj)
        strings = [AXObject.get_name(label) or AXText.get_all_text(label) for label in labels]
        result = " ".join(strings)
        return result

    @staticmethod
    def get_displayed_description(obj: Atspi.Accessible) -> str:
        """Returns the displayed description of obj."""

        descriptions = AXUtilitiesRelation.get_is_described_by(obj)
        strings = [AXObject.get_name(desc) or AXText.get_all_text(desc) for desc in descriptions]
        result = " ".join(strings)
        return result

    @staticmethod
    def get_heading_level(obj: Atspi.Accessible) -> int:
        """Returns the heading level of obj."""

        if not AXUtilitiesRole.is_heading(obj):
            return 0

        use_cache = not AXUtilitiesState.is_editable(obj)
        attrs = AXObject.get_attributes_dict(obj, use_cache)

        try:
            value = int(attrs.get("level", "0"))
        except ValueError:
            tokens = ["AXUtilities: Exception getting value for", obj, "(", attrs, ")"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return 0

        return value

    @staticmethod
    def get_nesting_level(obj: Atspi.Accessible) -> int:
        """Returns the nesting level of obj."""

        def pred(x: Atspi.Accessible) -> bool:
            if AXUtilitiesRole.is_list_item(obj):
                return AXUtilitiesRole.is_list(AXObject.get_parent(x))
            return AXUtilitiesRole.have_same_role(obj, x)

        ancestors = []
        ancestor = AXObject.find_ancestor(obj, pred)
        while ancestor:
            ancestors.append(ancestor)
            ancestor = AXObject.find_ancestor(ancestor, pred)

        return len(ancestors)

    @staticmethod
    def get_next_object(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the next object (depth first, unless there's a flows-to relation)"""

        if not AXObject.is_valid(obj):
            return None

        targets = AXUtilitiesRelation.get_flows_to(obj)
        for target in targets:
            if not AXObject.is_dead(target):
                return target

        index = AXObject.get_index_in_parent(obj) + 1
        parent = AXObject.get_parent(obj)
        while parent and not 0 < index < AXObject.get_child_count(parent):
            obj = parent
            index = AXObject.get_index_in_parent(obj) + 1
            parent = AXObject.get_parent(obj)

        if parent is None:
            return None

        next_object = AXObject.get_child(parent, index)
        if next_object == obj:
            tokens = ["AXUtilities:", obj, "claims to be its own next object"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return next_object

    @staticmethod
    def get_previous_object(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the previous object (depth first, unless there's a flows-from relation)"""

        if not AXObject.is_valid(obj):
            return None

        targets = AXUtilitiesRelation.get_flows_from(obj)
        for target in targets:
            if not AXObject.is_dead(target):
                return target

        index = AXObject.get_index_in_parent(obj) - 1
        parent = AXObject.get_parent(obj)
        while parent and not 0 <= index < AXObject.get_child_count(parent) - 1:
            obj = parent
            index = AXObject.get_index_in_parent(obj) - 1
            parent = AXObject.get_parent(obj)

        if parent is None:
            return None

        previous_object = AXObject.get_child(parent, index)
        if previous_object == obj:
            tokens = ["AXUtilities:", obj, "claims to be its own previous object"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return previous_object

    @staticmethod
    def is_on_screen(
        obj: Atspi.Accessible,
        bounding_box: Atspi.Rect | None = None
    ) -> bool:
        """Returns true if obj should be treated as being on screen."""

        AXObject.clear_cache(obj, False, "Updating to check if object is on screen.")

        tokens = ["AXUtilities: Checking if", obj, "is showing and visible...."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not (AXUtilitiesState.is_showing(obj) and AXUtilitiesState.is_visible(obj)):
            tokens = ["AXUtilities:", obj, "is not showing and visible. Treating as off screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXUtilities:", obj, "is showing and visible. Checking hidden..."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if AXUtilitiesState.is_hidden(obj):
            tokens = ["AXUtilities:", obj, "is reports being hidden. Treating as off screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXUtilities:", obj, "is not hidden. Checking size and rect..."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if AXComponent.has_no_size_or_invalid_rect(obj):
            tokens = ["AXUtilities: Rect of", obj, "is unhelpful. Treating as on screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        tokens = ["AXUtilities:", obj, "has size and a valid rect. Checking if off screen..."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if AXComponent.object_is_off_screen(obj):
            tokens = ["AXUtilities:", obj, "is believed to be off screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXUtilities:", obj, "is not off screen. Checking",
                  bounding_box, "intersection..."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if bounding_box is not None and not AXComponent.object_intersects_rect(obj, bounding_box):
            tokens = ["AXUtilities", obj, "not in", bounding_box, ". Treating as off screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXUtilities:", obj, "is believed to be on screen."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    @staticmethod
    def treat_as_leaf_node(obj: Atspi.Accessible) -> bool:
        """Returns True if obj should be treated as a leaf node."""

        if AXUtilitiesRole.children_are_presentational(obj):
            # In GTK, the contents of the page tab descends from the page tab.
            if AXUtilitiesRole.is_page_tab(obj):
                return False
            return True

        role = AXObject.get_role(obj)
        if AXUtilitiesRole.is_combo_box(obj, role):
            return not AXUtilitiesState.is_expanded(obj)

        if AXUtilitiesRole.is_menu(obj, role) and not AXUtilitiesRole.has_role_from_aria(obj):
            return not AXUtilitiesState.is_expanded(obj)

        if AXObject.get_name(obj):
            return AXUtilitiesRole.is_link(obj, role) or AXUtilitiesRole.is_label(obj, role)

        return False

    @staticmethod
    def _get_on_screen_objects(
        root: Atspi.Accessible,
        cancellation_event: threading.Event,
        bounding_box: Atspi.Rect | None = None
    ) -> list:

        tokens = ["AXUtilities: Getting on-screen objects in", root, f"({hex(id(root))})"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if cancellation_event.is_set():
            msg = "AXUtilities: Cancellation event set. Stopping search."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return []

        if not AXUtilities.is_on_screen(root, bounding_box):
            return []

        if AXUtilities.treat_as_leaf_node(root):
            return [root]

        if AXObject.supports_table(root) and AXObject.supports_selection(root):
            return list(AXTable.iter_visible_cells(root))

        objects = []
        root_name = AXObject.get_name(root)
        if root_name or AXObject.get_description(root) or AXText.has_presentable_text(root):
            objects.append(root)

        if bounding_box is None:
            bounding_box = AXComponent.get_rect(root)

        for i, child in enumerate(AXObject.iter_children(root)):
            tokens = [f"AXUtilities: Child {i} is", child, f"({hex(id(child))})"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            if cancellation_event.is_set():
                msg = "AXUtilities: Cancellation event set. Stopping search."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                break

            children = AXUtilities._get_on_screen_objects(child, cancellation_event, bounding_box)
            objects.extend(children)
            if root_name and children and root in objects and root_name == AXObject.get_name(child):
                objects.remove(root)

        if objects:
            return objects

        if AXUtilitiesState.is_focusable(root) or AXObject.has_action(root, "click"):
            return [root]

        return []

    @staticmethod
    def get_on_screen_objects(
        root: Atspi.Accessible,
        bounding_box: Atspi.Rect | None = None,
        timeout: float = 5.0
    ) -> list:
        """Returns a list of onscreen objects in the given root."""

        result_queue: queue.Queue[list] = queue.Queue()
        cancellation_event = threading.Event()

        def _worker():
            result = AXUtilities._get_on_screen_objects(root, cancellation_event, bounding_box)
            if not cancellation_event.is_set():
                result_queue.put(result)

        worker_thread = threading.Thread(target=_worker)
        worker_thread.start()

        try:
            result = result_queue.get(timeout=timeout)
        except queue.Empty:
            tokens = ["AXUtilities: get_on_screen_objects timed out.", root]
            debug.print_tokens(debug.LEVEL_WARNING, tokens, True)
            cancellation_event.set()
            result = []

            msg = "AXUtilities: Checking AT-SPI responsiveness...."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            desktop = AXUtilitiesApplication.get_desktop()
            tokens = ["AXUtilities: Desktop is", desktop]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        worker_thread.join()
        tokens = [f"AXUtilities: {len(result)} onscreen objects found in", root]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

for method_name, method in inspect.getmembers(AXUtilitiesApplication, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesEvent, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesRelation, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesRole, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesState, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesCollection, predicate=inspect.isfunction):
    if method_name.startswith("find"):
        setattr(AXUtilities, method_name, method)

AXUtilities.start_cache_clearing_thread()
